#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import timedelta
import functools
import warnings
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit, OptimizeWarning
from covsirphy.util.evaluator import Evaluator
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated.trend_plot import trend_plot


class _SRChange(Term):
    """
    Perform S-R trend analysis.

    Args:
        sr_df (pandas.DataFrame)
            Index
                Date (pd.TimeStamp): Observation date
            Columns
                - Recovered (int): the number of recovered cases (>0)
                - Susceptible (int): the number of susceptible cases
                - any other columns will be ignored
    """

    def __init__(self, sr_df):
        Validator(sr_df, name="sr_df").dataframe(time_index=True, columns=[self.S, self.R])
        # Index: Date, Columns: Recovered, Susceptible, logS
        self._sr_df = pd.DataFrame(
            {
                self.R: sr_df[self.R],
                self.S: sr_df[self.S],
                "logS": np.log10(sr_df[self.S].astype(np.float64)),
            }
        )

    def run(self, algorithm, **kwargs):
        """
        Run optimization and return the change points.

        Args:
            algorithm (classes of ruptures): detection algorithms
            kwargs: the other arguments of the algorithm class

        Returns:
            list[pandas.Timestamp]: list of change points
        """
        # Index: Recovered, Columns: logS
        df = self._sr_df.pivot_table(index=self.R, values="logS", aggfunc="last")
        df.index.name = None
        # Detect change points with Ruptures package: reset index + 1 values will be returned
        results = algorithm.fit_predict(df.iloc[:, 0].to_numpy(), pen=0.5)[:-1]
        # Convert reset index + 1 values to logS
        logs_df = df.iloc[[result - 1 for result in results]]
        # Convert logS to dates
        merge_df = pd.merge_asof(
            logs_df.sort_values("logS"), self._sr_df.reset_index().sort_values("logS"),
            on="logS", direction="nearest")
        return merge_df[self.DATE].sort_values().tolist()

    def _fitting_in_phase(self, start, end):
        """
        Perform curve fitting with the actual values in a phase on S-R plane.

        Args:
            start (pandas.Timestamp): start date of the phase
            end (pandas.Timestamp): end date of the phase
            func (function): function for fitting

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): observation date
                Columns
                    Fitted (float): fitted values of Susceptible
        """
        # Actual values for the phase
        df = self._sr_df.loc[start: end]
        # Curve fitting
        warnings.filterwarnings("ignore", category=OptimizeWarning)
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        param, _ = curve_fit(self._linear, df[self.R], df["logS"], maxfev=10000)
        # Get fitted values
        f_partial = functools.partial(self._linear, a=param[0], b=param[1])
        return pd.DataFrame({self.FITTED: 10 ** f_partial(df[self.R])}, index=df.index)

    @staticmethod
    def _linear(x, a, b):
        """
        Linear function f(x) = A x + b.

        Args:
            x (float): x values
            a (float): the first parameter of the function
            b (float): the second parameter of the function

        Returns:
            float
        """
        return a * x + b

    def _fitting(self, change_points):
        """
        Perform curve fitting with the actual values on S-R plane.

        Args:
            change_points (list[pandas.Timestamp]): list of change points

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): observation date
                Columns
                    - Recovered (int): The number of recovered cases
                    - Actual (int): actual values of Susceptible
                    - 0th, 1st, 2nd,... (float or None): fitted values of Susceptible for phases
        """
        df = self._sr_df.copy()
        # Start dates
        start_points = [df.index.min(), *change_points]
        # End dates
        end_points = [point - timedelta(days=1) for point in change_points] + [df.index.max()]
        # Phase names
        phases = [self.num2str(num) for num in range(len(change_points) + 1)]
        # Fitting
        df = df.rename(columns={self.S: self.ACTUAL}).drop("logS", axis=1)
        for (phase, start, end) in zip(phases, start_points, end_points):
            phase_df = self._fitting_in_phase(start=start, end=end)
            df = df.join(phase_df.rename(columns={self.FITTED: phase}), how="left")
        return df.round(0)

    def score(self, change_points, metric):
        """
        Calculate scores of the phases.

        Args:
            change_points (list[pandas.Timestamp]): list of change points
            metric (str): metric name

        Returns:
            list[float]: scores for phases

        Note:
            Please refer to covsirphy.Evaluator.score() for metric names
        """
        fit_df = self._fitting(change_points)
        phases = [self.num2str(num) for num in range(len(change_points) + 1)]
        scores = []
        for phase in phases:
            df = fit_df[[self.ACTUAL, phase]].dropna()
            evaluator = Evaluator(df[self.ACTUAL], df[phase], how="all")
            scores.append(evaluator.score(metric=metric))
        return scores

    def show(self, change_points, area, **kwargs):
        """
        Show the trend on S-R plane.

        Args:
            change_points (list[pandas.Timestamp]): list of change points
            area (str): area name (used in the figure title)
            kwargs: keyword arguments of covsirphy.trend_plot()
        """
        # Title
        if change_points:
            title = f"{area}: phases detected with S-R trend analysis"
        else:
            title = f"{area}: S-R trend without change points"
        # Curve fitting
        fit_df = self._fitting(change_points)
        # Show S-R plane
        fit_df = fit_df.rename(columns={"0th": self.INITIAL})
        plot_kwargs = {
            "title": title,
            "xlabel": self.R,
            "ylabel": self.S,
            "show_legend": True,
            "v": [fit_df.loc[point, self.R] for point in change_points],
        }
        plot_kwargs.update(kwargs)
        trend_plot(df=fit_df.set_index(self.R), actual_col=self.ACTUAL, **plot_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from unidecode import unidecode
from covsirphy.util.term import Term
from covsirphy.downloading.db import _DataBase


class _GoogleOpenData(_DataBase):
    """
    Access "COVID-19 Open Data by Google Cloud Platform" server.
    https://github.com/GoogleCloudPlatform/covid-19-open-data
    https://goo.gle/covid-19-open-data
    """
    # File title without extensions and suffix
    TITLE = "google_cloud_platform"
    # Dictionary of column names
    _MOBILITY_COLS_RAW = [
        "mobility_grocery_and_pharmacy",
        "mobility_parks",
        "mobility_transit_stations",
        "mobility_retail_and_recreation",
        "mobility_residential",
        "mobility_workplaces",
    ]
    MOBILITY_VARS = [v.capitalize() for v in _MOBILITY_COLS_RAW]
    COL_DICT = {
        "date": Term.DATE,
        **dict(zip(_MOBILITY_COLS_RAW, MOBILITY_VARS)),
    }
    # Stdout when downloading (shown at most one time)
    STDOUT = "Retrieving datasets from COVID-19 Open Data by Google Cloud Platform https://github.com/GoogleCloudPlatform/covid-19-open-data"
    # Citation
    CITATION = "O. Wahltinez and others (2020)," \
        " COVID-19 Open-Data: curating a fine-grained, global-scale data repository for SARS-CoV-2, " \
        " Work in progress, https://goo.gle/covid-19-open-data"

    def _country(self):
        """Returns country-level data.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (pandas.NA): NAs
                    - City (pandas.NA): NAs
                    - Mobility_grocery_and_pharmacy: % to baseline in visits (grocery markets, pharmacies etc.)
                    - Mobility_parks: % to baseline in visits (parks etc.)
                    - Mobility_transit_stations: % to baseline in visits (public transport hubs etc.)
                    - Mobility_retail_and_recreation: % to baseline in visits (restaurant, museums etc.)
                    - Mobility_residential: % to baseline in visits (places of residence)
                    - Mobility_workplaces: % to baseline in visits (places of work)
        """
        level_file = self._filer.csv(title=f"{self.TITLE}_{self.COUNTRY}".lower())["path_or_buf"]
        if self._provider.download_necessity(level_file):
            df = self._mobility()
            df = df.loc[df["ISO2"].str.len() == 2]
            df = df.merge(self._country_information()[["ISO2", "ISO3"]], how="left", on="ISO2")
            df = df.drop("ISO2", axis=1).dropna(subset=["ISO3"]).rename(columns={"ISO3": self.ISO3})
            df.to_csv(level_file, index=False)
        else:
            df = self._provider.read_csv(level_file, columns=None, date=self.DATE, date_format="%Y-%m-%d")
        df[self.PROVINCE] = pd.NA
        df[self.CITY] = pd.NA
        return df

    def _province(self, country):
        """Returns province-level data.

        Args:
            country (str): country name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (str): province/state/prefecture names
                    - City (pandas.NA): NAs
                    - Mobility_grocery_and_pharmacy: % to baseline in visits (grocery markets, pharmacies etc.)
                    - Mobility_parks: % to baseline in visits (parks etc.)
                    - Mobility_transit_stations: % to baseline in visits (public transport hubs etc.)
                    - Mobility_retail_and_recreation: % to baseline in visits (restaurant, museums etc.)
                    - Mobility_residential: % to baseline in visits (places of residence)
                    - Mobility_workplaces: % to baseline in visits (places of work)
        """
        iso3 = self._to_iso3(country)[0]
        index_df = self._index_data()
        # Mobility data
        level_file = self._filer.csv(title=f"{self.TITLE}_{self.PROVINCE}".lower())["path_or_buf"]
        if self._provider.download_necessity(level_file):
            df = self._mobility()
            df = df.loc[df["ISO2"].str.len() != 2]
            df = df.merge(index_df, how="left", on="ISO2")
            df = df.drop("ISO2", axis=1).dropna(subset=["ISO3"]).rename(columns={"ISO3": self.ISO3})
            df = df.loc[(df[self.PROVINCE] != self.NA) & (df[self.CITY] == self.NA)]
            df.to_csv(level_file, index=False)
        else:
            df = self._provider.read_csv(level_file, columns=None, date=self.DATE, date_format="%Y-%m-%d")
        return df.loc[df[self.ISO3] == iso3]

    def _city(self, country, province):
        """Returns city-level data.

        Args:
            country (str): country name
            province (str): province/state/prefecture name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (str): province/state/prefecture names
                    - City (str): city names
                    - Mobility_grocery_and_pharmacy: % to baseline in visits (grocery markets, pharmacies etc.)
                    - Mobility_parks: % to baseline in visits (parks etc.)
                    - Mobility_transit_stations: % to baseline in visits (public transport hubs etc.)
                    - Mobility_retail_and_recreation: % to baseline in visits (restaurant, museums etc.)
                    - Mobility_residential: % to baseline in visits (places of residence)
                    - Mobility_workplaces: % to baseline in visits (places of work)
        """
        iso3 = self._to_iso3(country)[0]
        index_df = self._index_data()
        # Mobility data
        level_file = self._filer.csv(title=f"{self.TITLE}_{self.CITY}".lower())["path_or_buf"]
        if self._provider.download_necessity(level_file):
            df = self._mobility()
            df = df.loc[df["ISO2"].str.len() != 2]
            df = df.merge(index_df, how="left", on="ISO2")
            df = df.drop("ISO2", axis=1).dropna(subset=["ISO3"]).rename(columns={"ISO3": self.ISO3})
            df = df.loc[df[self.CITY] != self.NA]
            df.to_csv(level_file, index=False)
        else:
            df = self._provider.read_csv(level_file, columns=None, date=self.DATE, date_format="%Y-%m-%d")
        return df.loc[(df[self.ISO3] == iso3) & (df[self.PROVINCE] == province)]

    def _index_data(self):
        """Returns index data.

        Args:
            country (str): country name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - location key (str): Google location keys
                    - ISO2 (str): ISO2 codes
                    - Province (str): province/state/prefecture names
                    - City (str): city names
        """
        URL_I = "https://storage.googleapis.com/covid19-open-data/v3/index.csv"
        index_cols = ["location_key", "subregion1_name", "subregion2_name", "locality_name", "iso_3166_1_alpha_3"]
        df = self._provide(url=URL_I, suffix="_index", columns=index_cols, date=None, date_format=None)
        df = df.rename(columns={"subregion1_name": self.PROVINCE, "iso_3166_1_alpha_3": self.ISO3})
        df[self.PROVINCE] = df[self.PROVINCE].fillna(self.NA).apply(unidecode)
        df[self.CITY] = df["subregion2_name"].fillna(df["locality_name"]).fillna(self.NA).apply(unidecode)
        return df.fillna(self.NA).loc[:, ["location_key", self.ISO3, self.PROVINCE, self.CITY]].rename(columns={"location_key": "ISO2"})

    def _mobility(self):
        """Returns mobility data.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO2 (str): ISO2 codes
                    - Mobility_grocery_and_pharmacy: % to baseline in visits (grocery markets, pharmacies etc.)
                    - Mobility_parks: % to baseline in visits (parks etc.)
                    - Mobility_transit_stations: % to baseline in visits (public transport hubs etc.)
                    - Mobility_retail_and_recreation: % to baseline in visits (restaurant, museums etc.)
                    - Mobility_residential: % to baseline in visits (places of residence)
                    - Mobility_workplaces: % to baseline in visits (places of work)
        """
        URL_M = "https://storage.googleapis.com/covid19-open-data/v3/mobility.csv"
        df = self._provide(
            url=URL_M, suffix="", columns=["date", "location_key", *self._MOBILITY_COLS_RAW], date="date", date_format="%Y-%m-%d")
        df.loc[:, self.MOBILITY_VARS] = df.loc[:, self.MOBILITY_VARS] + 100
        return df.rename(columns={"location_key": "ISO2"})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
import numpy as np
import pandas as pd
from covsirphy.util.error import SubsetNotFoundError, deprecate
from covsirphy.util.validator import Validator
from covsirphy.cleaning.cbase import CleaningBase
from covsirphy._deprecated.country_data import CountryData
from covsirphy._deprecated.jhu_complement import JHUDataComplementHandler


class JHUData(CleaningBase):
    """
    Data cleaning of JHU-style dataset.

    Args:
        filename (str or None): CSV filename of the dataset
        data (pandas.DataFrame or None):
            Index
                reset index
            Columns
                - Date: Observation date
                - ISO3: ISO3 code
                - Country: country/region name
                - Province: province/prefecture/state name
                - Confirmed: the number of confirmed cases
                - Fatal: the number of fatal cases
                - Recovered: the number of recovered cases
                - Population: population values
        citation (str or None): citation or None (empty)

    Note:
        Either @filename (high priority) or @data must be specified.

    Note:
        The number of infected cases will be (re-)calculated when data cleaning automatically.
    """
    # For JHUData.from_dataframe()
    _RAW_COLS_DEFAULT = [
        CleaningBase.DATE, CleaningBase.ISO3, CleaningBase.COUNTRY, CleaningBase.PROVINCE,
        CleaningBase.C, CleaningBase.CI, CleaningBase.F, CleaningBase.R, CleaningBase.N
    ]

    @deprecate(old="JHUData", new="DataEngineer", version="2.24.0-xi")
    def __init__(self, filename=None, data=None, citation=None):
        variables = [self.C, self.CI, self.F, self.R, self.N]
        super().__init__(filename=filename, data=data, citation=citation, variables=variables)
        # Recovery period
        self._recovery_period = None

    @property
    def recovery_period(self):
        """
        int: expected value of recovery period [days]
        """
        self._recovery_period = self._recovery_period or self.calculate_recovery_period()
        return self._recovery_period

    @recovery_period.setter
    def recovery_period(self, value):
        self._recovery_period = self._ensure_natural_int(value)

    def cleaned(self, **kwargs):
        """
        Return the cleaned dataset.

        Args:
            kwargs: keyword arguments will be ignored

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): Observation date
                    - ISO3: ISO3 code
                    - Country (pandas.Category): country/region name
                    - Province (pandas.Category): province/prefecture/state name
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - Population: population values
        """
        if "population" in kwargs.keys():
            raise ValueError(
                "@population was removed in JHUData.cleaned(). Please use JHUData.subset()")
        df = self._cleaned_df.copy()
        df[self.CI] = (df[self.C] - df[self.F] - df[self.R]).astype(np.int64)
        return df

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - ISO3 (str): ISO3 code
                    - Country (pandas.Category): country/region name
                    - Province (pandas.Category): province/prefecture/state name
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - Population (int): population values or 0 (when raw data is 0 values)
        """
        df = self._raw.loc[:, self._raw_cols]
        # Datetime columns
        df[self.DATE] = pd.to_datetime(df[self.DATE]).dt.round("D")
        with contextlib.suppress(TypeError):
            df[self.DATE] = df[self.DATE].dt.tz_convert(None)
        # Province
        df[self.PROVINCE] = df[self.PROVINCE].fillna(self.NA)
        # Values
        for col in [self.C, self.F, self.R, self.N]:
            df[col] = df.groupby([self.COUNTRY, self.PROVINCE])[col].ffill().fillna(0).astype(np.int64)
        # Calculate Infected
        df[self.CI] = (df[self.C] - df[self.F] - df[self.R]).astype(np.int64)
        # Update data types to reduce memory
        df[self.AREA_ABBR_COLS] = df[self.AREA_ABBR_COLS].astype("category")
        return df.loc[:, self._raw_cols]

    @deprecate("JHUData.replace()", version="2.21.0-xi-fu1")
    def replace(self, country_data):
        """
        Replace a part of cleaned dataset with a dataframe.

        Args:
            country_data (covsirphy.CountryData): dataset object of the country

        Returns:
            covsirphy.JHUData: self

        Note:
            Citation of the country data will be added to 'JHUData.citation' description.
        """
        self._ensure_instance(country_data, CountryData, name="country_data")
        df = self._cleaned_df.copy()
        # Read new dataset
        country = country_data.country
        new = country_data.cleaned()
        new[self.ISO3] = self.country_to_iso3(country)
        # Add population data
        new[self.N] = new.loc[:, self.N] if self.N in new else None
        new = new.set_index([self.COUNTRY, self.PROVINCE, self.DATE])
        new.update(df.set_index([self.COUNTRY, self.PROVINCE, self.DATE]).loc[:, [self.N]])
        new = new.reset_index().loc[:, self._raw_cols]
        # Calculate Infected
        new[self.CI] = (new[self.C] - new[self.F] - new[self.R]).astype(np.int64)
        # Remove the data in the country from JHU dataset
        df = df.loc[df[self.COUNTRY] != country]
        # Combine JHU data and the new data
        df = pd.concat([df, new], axis=0, sort=False)
        # Update data types to reduce memory
        df[self.AREA_ABBR_COLS] = df[self.AREA_ABBR_COLS].astype("category")
        self._cleaned_df = df.copy()
        # Citation
        self._citation += f"\n{country_data.citation}"
        return self

    def _calculate_susceptible(self, subset_df, population):
        """
        Return the subset of dataset.

        Args:
            subset_df (pandas.DataFrame)
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - Population (int): population values or 0 values (0 will be ignored)
            population (int or None): population value

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - Susceptible (int): the number of susceptible cases, if calculated

        Note:
            If @population (high priority) is not None or population values are registered in subset,
            the number of susceptible cases will be calculated.
        """
        df = subset_df.copy()
        df[self.S] = (population or df[self.N]) - df[self.C]
        try:
            df[self.S] = df[self.S].astype(np.int64)
        except ValueError:
            return df.loc[:, [self.DATE, self.C, self.CI, self.F, self.R]]
        return df.loc[:, [self.DATE, self.C, self.CI, self.F, self.R, self.S]]

    def subset(self, country, province=None, start_date=None, end_date=None,
               population=None, recovered_min=1):
        """
        Return the subset of dataset.

        Args:
            country (str): country name or ISO3 code
            province (str or None): province name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            population (int or None): population value
            recovered_min (int): minimum number of recovered cases records must have

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases (> 0)
                    - Susceptible (int): the number of susceptible cases, if calculated

        Note:
            If @population (high priority) is not None or population values are registered in subset,
            the number of susceptible cases will be calculated.
        """
        country_alias = self.ensure_country_name(country)
        # Subset with area, start/end date
        try:
            subset_df = super().subset(
                country=country, province=province, start_date=start_date, end_date=end_date)
        except SubsetNotFoundError:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province,
                start_date=start_date, end_date=end_date) from None
        # Calculate Susceptible
        df = self._calculate_susceptible(subset_df, population)
        # Select records where Recovered >= recovered_min
        recovered_min = self._ensure_natural_int(recovered_min, name="recovered_min", include_zero=True)
        df = df.loc[df[self.R] >= recovered_min, :].reset_index(drop=True)
        if df.empty:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province,
                start_date=start_date, end_date=end_date,
                details=f"with 'Recovered >= {recovered_min}'") from None
        return df

    @deprecate("JHUData.to_sr()", version="2.17.0-zeta")
    def to_sr(self, country, province=None,
              start_date=None, end_date=None, population=None):
        """
        Create Susceptible/Recovered dataset without complement.

        Args:
            country (str): country name
            province (str): province name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            population (int): population value

        Returns:
            pandas.DataFrame
                Index
                    Date (pd.Timestamp): Observation date
                Columns
                    - Recovered (int): the number of recovered cases (> 0)
                    - Susceptible (int): the number of susceptible cases

        Note:
            @population must be specified.
            Records with Recovered > 0 will be used.
        """
        population = Validator(population, "population").int(value_range=(1, None))
        subset_df = self.subset(
            country=country, province=province,
            start_date=start_date, end_date=end_date, population=population)
        return subset_df.set_index(self.DATE).loc[:, [self.R, self.S]]

    @classmethod
    @deprecate("JHUData.from_dataframe()", new="DataLoader.read_dataframe()",
               version="2.21.0-xi-fu1", ref="https://lisphilar.github.io/covid19-sir/markdown/LOADING.html")
    def from_dataframe(cls, dataframe, directory="input"):
        """
        Create JHUData instance using a pandas dataframe.

        Args:
            dataframe (pd.DataFrame): cleaned dataset
                Index
                    reset index
                Columns
                    - Date: Observation date
                    - ISO3: ISO3 code (optional)
                    - Country: country/region name
                    - Province: province/prefecture/state name
                    - Confirmed: the number of confirmed cases
                    - Infected: the number of currently infected cases
                    - Fatal: the number of fatal cases
                    - Recovered: the number of recovered cases
                    - Population: population values (optional)
            directory (str): directory to save geography information (for .map() method)

        Returns:
            covsirphy.JHUData: JHU-style dataset
        """
        df = cls._ensure_dataframe(dataframe, name="dataframe")
        df[cls.ISO3] = df[cls.ISO3] if cls.ISO3 in df else cls.NA
        df[cls.N] = df[cls.N] if cls.N in df else 0
        instance = cls()
        instance.directory = str(directory)
        instance._cleaned_df = cls._ensure_dataframe(df, name="dataframe", columns=cls._RAW_COLS_DEFAULT)
        return instance

    def total(self):
        """
        Calculate total number of cases and rates.

        Returns:
            pandas.DataFrame: group-by Date, sum of the values

                Index
                    Date (pandas.Timestamp): Observation date
                Columns
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - Fatal per Confirmed (int)
                    - Recovered per Confirmed (int)
                    - Fatal per (Fatal or Recovered) (int)
        """
        df = self._cleaned_df.copy()
        df = df.loc[df[self.PROVINCE] == self.NA]
        df = df.groupby(self.DATE).sum()
        total_series = df.loc[:, self.C]
        r_cols = self.RATE_COLUMNS[:]
        df[r_cols[0]] = df[self.F] / total_series
        df[r_cols[1]] = df[self.R] / total_series
        df[r_cols[2]] = df[self.F] / (df[self.F] + df[self.R])
        # Set the final date of the records
        raw_df = self._raw.copy()
        final_date = pd.to_datetime(raw_df[self.DATE]).dt.date.max()
        df = df.loc[df.index.date <= final_date]
        return df.loc[:, [*self.VALUE_COLUMNS, *r_cols]]

    def countries(self, complement=True, **kwargs):
        """
        Return names of countries where records.

        Args:
            complement (bool): whether say OK for complement or not
            interval (int): expected update interval of the number of recovered cases [days]
            kwargs: the other keyword arguments of JHUData.subset_complement()

        Returns:
            list[str]: list of country names
        """
        df = self._cleaned_df.copy()
        df = df.loc[df[self.PROVINCE] == self.NA]
        # All countries
        all_set = set((df[self.COUNTRY].unique()))
        # Selectable countries without complement
        raw_ok_set = set(df.loc[df[self.R] > 0, self.COUNTRY].unique())
        if not complement:
            return sorted(raw_ok_set)
        # Selectable countries
        comp_ok_list = []
        for country in all_set - raw_ok_set:
            with contextlib.suppress(SubsetNotFoundError):
                self.subset_complement(country=country, **kwargs)
                comp_ok_list.append(country)
        return sorted(raw_ok_set | set(comp_ok_list))

    @deprecate("JHUData.calculate_closing_period()")
    def calculate_closing_period(self):
        """
        Calculate mode value of closing period, time from confirmation to get outcome.

        Returns:
            int: closing period [days]

        Note:
            If no records we can use for calculation were registered, 12 [days] will be applied.
        """
        # Get cleaned dataset at country level
        df = self._cleaned_df.copy()
        df = df.loc[df[self.PROVINCE] == self.NA]
        # Select records of countries where recovered values are reported
        df = df.groupby(self.COUNTRY).filter(lambda x: x[self.R].sum() != 0)
        if df.empty:
            return 12
        # Total number of confirmed/closed cases of selected records
        df = df.groupby(self.DATE).sum()
        df[self.FR] = df[[self.F, self.R]].sum(axis=1)
        df = df.loc[:, [self.C, self.FR]]
        # Calculate how many days to confirmed, closed
        df = df.unstack().reset_index()
        df.columns = ["Variable", self.DATE, "Number"]
        df["Days"] = (df[self.DATE] - df[self.DATE].min()).dt.days
        df = df.pivot_table(values="Days", index="Number", columns="Variable")
        df = df.interpolate(limit_direction="both").fillna(method="ffill")
        df["Elapsed"] = df[self.FR] - df[self.C]
        df = df.loc[df["Elapsed"] > 0]
        # Calculate mode value of closing period
        return int(df["Elapsed"].mode().astype(np.int64).values[0])

    def calculate_recovery_period(self):
        """
        Calculate the median value of recovery period of all countries
        where recovered values are reported.

        Returns:
            int: recovery period [days]

        Note:
            If no records we can use for calculation were registered, 17 [days] will be applied.
        """
        default = 17
        # Get valid data for calculation
        df = self._cleaned_df.copy()
        df = df.loc[df[self.PROVINCE] == self.NA]
        df = df.groupby(self.COUNTRY).filter(lambda x: x[self.R].sum() != 0)
        # If no records were found the default value will be returned
        if df.empty:
            return default
        # Calculate median value of recovery period in all countries with valid data
        periods = [
            self._calculate_recovery_period_country(df, country)
            for country in df[self.COUNTRY].unique()
        ]
        valid_periods = list(filter(lambda x: x >= 0, periods))
        if not valid_periods:
            return default
        try:
            return int(pd.Series(valid_periods).median())
        except ValueError:
            return default

    def _calculate_recovery_period_country(self, valid_df, country, upper_limit_days=90,
                                           lower_limit_days=7, upper_percentage=0.5, lower_percentage=0.5):
        """
        Calculate mode value of recovery period in the country.
        If many mode values were found, mean value of mode values will be returned.

        Args:
            valid_df (pandas.DataFrame):
                Index
                    reset_index
                Columns
                    Date, Confirmed, Recovered, Fatal
            country(str): country name or ISO3 code
            upper_limit_days (int): maximum number of valid partial recovery periods [days]
            lower_limit_days (int): minimum number of valid partial recovery periods [days]
            upper_percentage (float): fraction of partial recovery periods with value greater than upper_limit_days
            lower_percentage (float): fraction of partial recovery periods with value less than lower_limit_days

        Returns:
            int: mode value of recovery period [days]
        """
        # Select country data
        df = valid_df.copy()
        df = df.loc[df[self.COUNTRY] == country].groupby(self.DATE).sum()
        # Calculate "Confirmed - Fatal"
        df["diff"] = df[self.C] - df[self.F]
        df = df.loc[:, ["diff", self.R]]
        # Calculate how many days passed to reach the number of cases
        df = df.unstack().reset_index()
        df.columns = ["Variable", "Date", "Number"]
        df["Days"] = (df[self.DATE] - df[self.DATE].min()).dt.days
        # Calculate recovery period (mode value because bimodal)
        df = df.pivot_table(values="Days", index="Number", columns="Variable")
        df = df.interpolate(limit_area="inside").dropna().astype(np.int64)
        df["Elapsed"] = df[self.R] - df["diff"]
        df = df.loc[df["Elapsed"] > 0]
        # Check partial recovery periods
        per_up = (df["Elapsed"] > upper_limit_days).sum()
        per_lw = (df["Elapsed"] < lower_limit_days).sum()
        if df.empty or per_up / len(df) >= upper_percentage or per_lw / len(df) >= lower_percentage:
            return -1
        return df["Elapsed"].mode().mean()

    def subset_complement(self, country, province=None,
                          start_date=None, end_date=None, population=None, **kwargs):
        """
        Return the subset of dataset and complement recovered data, if necessary.
        Records with Recovered > 0 will be selected.

        Args:
            country(str): country name or ISO3 code
            province(str or None): province name
            start_date(str or None): start date, like 22Jan2020
            end_date(str or None): end date, like 01Feb2020
            population(int or None): population value
            kwargs: keyword arguments of JHUDataComplementHandler(), control factors of complement

        Returns:
            tuple(pandas.DataFrame, str or bool):
                pandas.DataFrame:
                    Index
                        reset index
                    Columns
                        - Date(pd.Timestamp): Observation date
                        - Confirmed(int): the number of confirmed cases
                        - Infected(int): the number of currently infected cases
                        - Fatal(int): the number of fatal cases
                        - Recovered (int): the number of recovered cases ( > 0)
                        - Susceptible(int): the number of susceptible cases, if calculated
                str or bool: kind of complement or False

        Note:
            If @population (high priority) is not None or population values are registered in subset,
            the number of susceptible cases will be calculated.
        """
        # Subset with area
        country_alias = self.ensure_country_name(country)
        try:
            subset_df = super().subset(country=country, province=province, start_date=None, end_date=None)
        except SubsetNotFoundError:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province) from None
        # Complement, if necessary
        self._recovery_period = self._recovery_period or self.calculate_recovery_period()
        handler = JHUDataComplementHandler(recovery_period=self._recovery_period, **kwargs)
        df, status, _ = handler.run(subset_df)
        # Subsetting with dates
        if start_date is not None:
            df = df.loc[df[self.DATE] >= self._ensure_date(start_date, name="start_date")]
        if end_date is not None:
            df = df.loc[df[self.DATE] <= self._ensure_date(end_date, name="end_date")]
        if df.empty:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province,
                start_date=start_date, end_date=end_date) from None
        # Calculate Susceptible
        df.loc[:, self.N] = subset_df.loc[:, self.N]
        df = self._calculate_susceptible(df, population)
        # Kind of complement or False
        is_complemented = status or False
        # Select records where Recovered > 0
        df = df.loc[df[self.R] > 0, :].reset_index(drop=True)
        return (df, is_complemented)

    def records(self, country, province=None, start_date=None, end_date=None, population=None,
                auto_complement=True, **kwargs):
        """
        JHU-style dataset for the area from the start date to the end date.
        Records with Recovered > 0 will be selected.

        Args:
            country(str): country name or ISO3 code
            province(str or None): province name
            start_date(str or None): start date, like 22Jan2020
            end_date(str or None): end date, like 01Feb2020
            population(int or None): population value
            auto_complement (bool): if True and necessary, the number of cases will be complemented
            kwargs: the other arguments of JHUData.subset_complement()

        Raises:
            SubsetNotFoundError: failed in subsetting because of lack of data

        Returns:
            tuple(pandas.DataFrame, bool):
                pandas.DataFrame:

                    Index
                        reset index
                    Columns
                        - Date(pd.Timestamp): Observation date
                        - Confirmed(int): the number of confirmed cases
                        - Infected(int): the number of currently infected cases
                        - Fatal(int): the number of fatal cases
                        - Recovered (int): the number of recovered cases ( > 0)
                        - Susceptible(int): the number of susceptible cases, if calculated
                str or bool: kind of complement or False

        Note:
            If @population (high priority) is not None or population values are registered in subset,
            the number of susceptible cases will be calculated.

        Note:
            If necessary and @auto_complement is True, complement recovered data.
        """
        country_alias = self.ensure_country_name(country)
        subset_arg_dict = {
            "country": country, "province": province,
            "start_date": start_date, "end_date": end_date, "population": population,
        }
        if auto_complement:
            df, is_complemented = self.subset_complement(**subset_arg_dict, **kwargs)
            if not df.empty:
                return (df, is_complemented)
        try:
            return (self.subset(**subset_arg_dict), False)
        except ValueError:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province,
                start_date=start_date, end_date=end_date, details="with 'Recovered > 0'") from None

    def show_complement(self, country=None, province=None,
                        start_date=None, end_date=None, **kwargs):
        """
        To monitor effectivity and safety of complement on JHU subset,
        we need to know what kind of complement was done for JHU subset
        for each country (if country/countries specified) or for all countries.

        Args:
            country (str or list[str] or None): country/countries name or None (all countries)
            province(str or None): province name
            start_date(str or None): start date, like 22Jan2020
            end_date(str or None): end date, like 01Feb2020
            kwargs: keyword arguments of JHUDataComplementHandler(), control factors of complement

        Raises:
            ValueError: @province was specified when @country is not a string
            covsirphy.SubsetNotFoundError: No records were registered for the area/dates

        Returns:
            pandas.DataFrame

                Index
                    reset index
                Columns
                    - country (str): country name
                    - province (str): province name
                    - Monotonic_confirmed (bool): True if applied for confirmed cases or False otherwise
                    - Monotonic_fatal (bool): True if applied for fatal cases or False otherwise
                    - Monotonic_recovered (bool): True if applied for recovered or False otherwise
                    - Full_recovered (bool): True if applied for recovered or False otherwise
                    - Partial_recovered (bool): True if applied for recovered or False otherwise
        """
        self._recovery_period = self._recovery_period or self.calculate_recovery_period()
        # Area name
        if country is None:
            country = [c for c in self._cleaned_df[self.COUNTRY].unique() if c != "Others"]
        province = province or self.NA
        if not isinstance(country, str) and province != self.NA:
            raise ValueError("@province cannot be specified when @country is not a string.")
        if not isinstance(country, list):
            country = [country]
        # Create complement handler
        handler = JHUDataComplementHandler(recovery_period=self._recovery_period, **kwargs)
        # Check each country
        complement_df = pd.DataFrame(
            columns=[self.COUNTRY, self.PROVINCE, *JHUDataComplementHandler.SHOW_COMPLEMENT_FULL_COLS])
        complement_df.set_index(self.COUNTRY, inplace=True)
        for cur_country in country:
            try:
                subset_df = super().subset(
                    country=cur_country, province=province, start_date=start_date, end_date=end_date)
            except SubsetNotFoundError:
                raise SubsetNotFoundError(
                    country=cur_country, province=province,
                    start_date=start_date, end_date=end_date) from None
            * _, complement_dict = handler.run(subset_df)
            complement_dict_values = pd.Series(complement_dict.values(), dtype=bool).values
            complement_df.loc[cur_country] = [province, *complement_dict_values]
        return complement_df.reset_index()

    def map(self, country=None, variable="Confirmed", date=None, **kwargs):
        """
        Create global colored map to show the values.

        Args:
            country (str or None): country name or None (global map)
            variable (str): variable name to show
            date (str or None): date of the records or None (the last value)
            kwargs: arguments of ColoredMap() and ColoredMap.plot()

        Note:
            When @country is None, country level data will be shown on global map.
            When @country is a country name, province level data will be shown on country map.
        """
        # Date
        date_str = date or self.cleaned()[self.DATE].max().strftime(self.DATE_FORMAT)
        country_str = country or "Global"
        title = f"{country_str}: the number of {variable.lower()} cases on {date_str}"
        # Global map
        if country is None:
            return self._colored_map_global(
                variable=variable, title=title, date=date, **kwargs)
        # Country-specific map
        return self._colored_map_country(
            country=country, variable=variable, title=title, date=date, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class _RateConverter(Term, BaseEstimator, TransformerMixin):
    """
    Convert Indicators(n) to Indicators(n)/Indicators(n-1).

    Args:
        to_convert (bool): whether convert the values or not

    Note:
        We can use this convert in scikit-learn pipeline because this is a child class of
        sklearn.base.BaseEstimator and sklearn.base.TransformerMixin.
    """

    def __init__(self, to_convert=True):
        self.to_convert = bool(to_convert)

    @property
    def to_convert_(self):
        """
        bool: whether convert the values or not
        """
        return self.to_convert

    def fit(self, X, y):
        """
        Return self because fitting is unnecessary.

        Args:
            X (pandas.DataFrame): input X samples
            y (pandas.DataFrame): input y samples

        Returns:
            _RateConverter: self
        """
        return self

    def transform(self, X):
        """
        Transform the data.

        Args:
            X (pandas.DataFrame): input samples

        Returns:
            pandas.DataFrame: transformed samples
        """
        Validator(X, "X").dataframe()
        if not self.to_convert:
            return X
        return X.div(X.shift(1)).replace(np.inf, 0).fillna(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from covsirphy.util.term import Term
from covsirphy.downloading.db import _DataBase


class _OWID(_DataBase):
    """
    Access "Our World In Data" server.
    https://github.com/owid/covid-19-data/tree/master/public/data
    https://ourworldindata.org/coronavirus
    """
    # File title without extensions and suffix
    TITLE = "ourworldindata"
    # Dictionary of column names
    COL_DICT = {
        # Vaccine
        "iso_code": Term.ISO3,
        "date": Term.DATE,
        "vaccines": Term.PRODUCT,
        "total_vaccinations": Term.VAC,
        "total_boosters": Term.VAC_BOOSTERS,
        "people_vaccinated": Term.V_ONCE,
        "people_fully_vaccinated": Term.V_FULL,
        # Tests
        "ISO code": Term.ISO3,
        "Date": Term.DATE,
        "Cumulative total": Term.TESTS,
    }
    # Stdout when downloading (shown at most one time)
    STDOUT = "Retrieving datasets from Our World In Data https://github.com/owid/covid-19-data/"
    # Citation
    CITATION = "Hasell, J., Mathieu, E., Beltekian, D. et al." \
        " A cross-country database of COVID-19 testing. Sci Data 7, 345 (2020)." \
        " https: //doi.org/10.1038/s41597-020-00688-8"

    def _country(self):
        """Returns country-level data.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (str): NAs
                    - City (str): NAs
                    - Tests (numpy.float64): the number of tests
                    - Product (numpy.int64): vaccine product names
                    - Vaccinations (numpy.int64): cumulative number of vaccinations
                    - Vaccinations_boosters (numpy.int64): cumulative number of booster vaccinations
                    - Vaccinated_once (numpy.int64): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (numpy.int64): cumulative number of people who received all doses prescrived by the protocol
        """
        # URLs
        URL_V = "https://raw.githubusercontent.com/owid/covid-19-data/master/public/data/vaccinations/"
        URL_V_REC = f"{URL_V}vaccinations.csv"
        URL_V_LOC = f"{URL_V}locations.csv"
        # URL for PCR data
        URL_P = "https://raw.githubusercontent.com/owid/covid-19-data/master/public/data/testing/"
        URL_P_REC = f"{URL_P}covid-testing-all-observations.csv"
        # Vaccine
        v_rec_cols = [
            "date", "iso_code", "total_vaccinations", "total_boosters", "people_vaccinated", "people_fully_vaccinated"]
        v_rec_df = self._provide(
            url=URL_V_REC, suffix="_vaccine", columns=v_rec_cols, date="date", date_format="%Y-%m-%d")
        v_loc_df = self._provide(
            url=URL_V_LOC, suffix="_vaccine_locations", columns=["iso_code", "vaccines"], date=None, date_format="%Y-%m-%d")
        v_df = v_rec_df.merge(v_loc_df, how="left", on=self.ISO3)
        # Tests
        pcr_rec_cols = ["ISO code", "Date", "Daily change in cumulative total", "Cumulative total"]
        pcr_df = self._provide(url=URL_P_REC, suffix="", columns=pcr_rec_cols, date="Date", date_format="%Y-%m-%d")
        pcr_df["cumsum"] = pcr_df.groupby(self.ISO3)["Daily change in cumulative total"].cumsum()
        pcr_df = pcr_df.assign(tests=lambda x: x[self.TESTS].fillna(x["cumsum"]))
        pcr_df.rename(columns={"tests": self.TESTS})
        pcr_df = pcr_df.loc[:, [self.ISO3, self.DATE, self.TESTS]]
        # Combine data (vaccinations/tests)
        df = v_df.merge(pcr_df, how="outer", on=[self.ISO3, self.DATE])
        # Location (iso3/province)
        df = df.loc[~df[self.ISO3].str.contains("OWID_")]
        df = df.loc[~df[self.ISO3].isna()]
        df[self.PROVINCE] = self.NA
        df[self.CITY] = self.NA
        return df

    def _province(self, country):
        """Returns province-level data.

        Args:
            country (str): country name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (str): province/state/prefecture names
                    - City (str): NAs
                    - Tests (numpy.float64): the number of tests
                    - Product (numpy.int64): vaccine product names
                    - Vaccinations (numpy.int64): cumulative number of vaccinations
                    - Vaccinations_boosters (numpy.int64): cumulative number of booster vaccinations
                    - Vaccinated_once (numpy.int64): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (numpy.int64): cumulative number of people who received all doses prescrived by the protocol
        """
        return self._empty()

    def _city(self, country, province):
        """Returns city-level data.

        Args:
            country (str): country name
            province (str): province/state/prefecture name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (str): province/state/prefecture names
                    - City (str): city names
                    - Tests (numpy.float64): the number of tests
                    - Product (numpy.int64): vaccine product names
                    - Vaccinations (numpy.int64): cumulative number of vaccinations
                    - Vaccinations_boosters (numpy.int64): cumulative number of booster vaccinations
                    - Vaccinated_once (numpy.int64): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (numpy.int64): cumulative number of people who received all doses prescrived by the protocol
        """
        return self._empty()

    def _empty(self):
        """Return empty dataframe.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (str): province/state/prefecture names
                    - City (str): city names
                    - Tests (numpy.float64): the number of tests
                    - Product (numpy.int64): vaccine product names
                    - Vaccinations (numpy.int64): cumulative number of vaccinations
                    - Vaccinations_boosters (numpy.int64): cumulative number of booster vaccinations
                    - Vaccinated_once (numpy.int64): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (numpy.int64): cumulative number of people who received all doses prescrived by the protocol
        """
        columns = [
            self.DATE, self.ISO3, self.PROVINCE, self.CITY, self.TESTS, self.PRODUCT, self.VAC, self.VAC_BOOSTERS, self.V_ONCE, self.V_FULL]
        return pd.DataFrame(columns=columns)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from datetime import timedelta
import matplotlib
from matplotlib import pyplot as plt
import pandas as pd
import ruptures as rpt
from scipy.optimize import curve_fit
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy.visualization.vbase import VisualizeBase


class _TrendAnalyzer(Term):
    """Class for S-R trend analysis to find change points of log10(S) - R of model specific variables.

    Args:
        data (pandas.DataFrame): new data to overwrite the current information
            Index
                Date (pandas.Timestamp): Observation date
            Columns
                Susceptible (int): the number of susceptible cases
                Infected (int): the number of currently infected cases
                Fatal (int): the number of fatal cases
                Recovered (int): the number of recovered cases
        model (covsirphy.ODEModel): definition of ODE model
        min_size (int): minimum value of phase length [days], be equal to or over 3

    Note:
        S-R trend analysis is original to Covsirphy, https://www.kaggle.com/code/lisphilar/covid-19-data-with-sir-model/notebook
        "Phase" means a sequential dates in which the parameters of SIR-derived models are fixed.
        "Change points" means the dates when trend was changed.
        "Change points" is the same as the start dates of phases except for the 0th phase.
    """

    def __init__(self, data, model, min_size):
        self._all_df = model.sr(data)
        self._algo_dict = {
            "Pelt-rbf": (rpt.Pelt, {"model": "rbf", "jump": 1, "min_size": min_size}),
            "Binseg-rbf": (rpt.Binseg, {"model": "rbf", "jump": 1, "min_size": min_size}),
            "Binseg-normal": (rpt.Binseg, {"model": "normal", "jump": 1, "min_size": min_size}),
            "BottomUp-rbf": (rpt.BottomUp, {"model": "rbf", "jump": 1, "min_size": min_size}),
            "BottomUp-normal": (rpt.BottomUp, {"model": "normal", "jump": 1, "min_size": min_size}),
        }
        self._r, self._logS = model._r, model._logS
        self._model_name = model._NAME

    def find_points(self, algo="Binseg-normal", **kwargs):
        """Find change points with S-R trend analysis.

        Args:
            algo (str): detection algorithms and models
            **kwargs: keyword arguments of algorithm classes (ruptures.Pelt, .Binseg, BottomUp) except for "model"

        Returns:
            tuple of (pandas.Timestamp): date of change points

        Note:
            Python library `ruptures` will be used for off-line change point detection.
            Refer to documentation of `ruptures` library, https://centre-borelli.github.io/ruptures-docs/
            Candidates of @algo are "Pelt-rbf", "Binseg-rbf", "Binseg-normal", "BottomUp-rbf", "BottomUp-normal".
        """
        algo_dict = self._algo_dict.copy()
        Validator([algo], "algo").sequence(candidates=algo_dict.keys())
        r, logS = self._r, self._logS
        sr_df = self._all_df.pivot_table(index=r, values=logS, aggfunc="last")
        sr_df.index.name = None
        detector = algo_dict[algo][0](**algo_dict[algo][1], **Validator(kwargs).kwargs(algo_dict[algo][0]))
        results = detector.fit_predict(sr_df.iloc[:, 0].to_numpy(), pen=0.5)[:-1]
        logs_df = sr_df.iloc[[result - 1 for result in results]]
        merge_df = pd.merge_asof(
            logs_df.sort_values(logS), self._all_df.reset_index().sort_values(logS), on=logS, direction="nearest")
        return merge_df[self.DATE].sort_values().tolist()

    def fitting(self, points):
        """Perform fitting of data segmented with change points with linear function.

        Args:
            points (tuple of (pandas.Timestamp)): date of change points

        Returns:
            pandas.Dataframe:
                Index
                    R (int): actual R (R of the ODE model) values
                Columns
                    Actual (float): actual log10(S) (common logarithm of S of the ODE model) values
                    Fitted (float): log10(S) values fitted with y = a * R + b
                    0th (float): log10(S) values fitted with y = a * R + b and 0th phase data
                    1st, 2nd... (float): fitted values of 1st, 2nd phases
        """
        r, logS = self._r, self._logS
        all_df = self._all_df.copy()
        starts = [all_df.index.min(), *points]
        ends = [point - timedelta(days=1) for point in points] + [all_df.index.max()]
        for i, (start, end) in enumerate(zip(starts, ends)):
            phase_df = all_df.loc[start: end, :]
            param, _ = curve_fit(self._linear_f, phase_df[r], phase_df[logS], maxfev=10000)
            all_df[self.num2str(i)] = self._linear_f(phase_df[r], a=param[0], b=param[1])
        all_df[self.FITTED] = all_df.drop([logS, r], axis=1).sum(axis=1)
        return all_df.rename(columns={logS: self.ACTUAL}).set_index(r).groupby(level=0).first()

    @staticmethod
    def _linear_f(x, a, b):
        """
        Linear function f(x) = A * x + b.

        Args:
            x (float): x values
            a (float): the first parameter of the function
            b (float): the second parameter of the function

        Returns:
            float
        """
        return a * x + b

    def display(self, points, fit_df, name, **kwargs):
        """Display data on log10(S) - R plane with phases.

        Args:
            points (tuple of (pandas.Timestamp)): date of change points
            fit_df (pandas.Dataframe):
                Index
                    R (int): actual R (R of the ODE model) values
                Columns
                    Actual (float): actual log10(S) (common logarithm of S of the ODE model) values
                    Fitted (float): log10(S) values fitted with y = a * R + b
                    0th (float): log10(S) values fitted with y = a * R + b and 0th phase data
                    1st, 2nd... (float): fitted values of 1st, 2nd phases
            name (str or None): name of dynamics to show in figures (e.g. "baseline") or None (un-set)
            **kwargs: keyword arguments of covsirphy.VisualizeBase() and matplotlib.legend.Legend()
        """
        with VisualizeBase(**Validator(kwargs, "keyword arguments").kwargs([VisualizeBase, plt.savefig])) as lp:
            # _, lp.ax = plt.subplots(1, 1)
            lp.ax.plot(
                fit_df.index, fit_df[self.ACTUAL], label=self.ACTUAL,
                color="black", marker=".", markeredgewidth=0, linewidth=0)
            for phase in fit_df.drop([self.ACTUAL, self.FITTED], axis=1).columns:
                lp.ax.plot(fit_df.index, fit_df[phase], label=phase)
            for r_value in [self._all_df.loc[point, self._r] for point in points]:
                lp.ax.axvline(x=r_value, color="black", linestyle=":")
            pre = "Phases" if name is None else name + ": phases"
            lp.title = f"{pre} detected by S-R trend analysis of {self._model_name}"
            lp.ax.set_xlim(max(0, min(fit_df.index)), None)
            lp.ax.set_xlabel(xlabel=f"R of {self._model_name}")
            lp.ax.set_ylabel(ylabel=f"log10(S) of {self._model_name}")
            legend_kwargs = Validator(kwargs, "keyword arguments").kwargs(
                matplotlib.legend.Legend, default=dict(bbox_to_anchor=(0.5, -0.5), loc="lower center", borderaxespad=0, ncol=7))
            lp.ax.legend(**legend_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from inspect import signature
from covsirphy.util.error import deprecate


@deprecate("find_args", version="2.25.0-mu")
def find_args(func_list, **kwargs):
    """
    Find values of enabled arguments of the function from the keyword arguments.

    Args:
        func_list (list[function] or function): target function
        kwargs: keyword arguments

    Returns:
        dict: dictionary of enabled arguments
    """
    if not isinstance(func_list, list):
        func_list = [func_list]
    enabled_nest = [
        list(signature(func).parameters.keys()) for func in func_list
    ]
    enabled_set = set(sum(enabled_nest, list()))
    enabled_set = enabled_set - {"self", "cls"}
    return {k: v for (k, v) in kwargs.items() if k in enabled_set}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from inspect import signature
import math
import pandas as pd
from covsirphy.util.error import NAFoundError, NotIncludedError, NotSubclassError, UnExpectedTypeError, EmptyError
from covsirphy.util.error import UnExpectedValueRangeError, UnExpectedValueError, UnExpectedLengthError, UnExpectedNoneError


class Validator(object):
    """Validate objects and arguments.

    Args:
        target (object): target object to validate
        name (str): name of the target shown in error code
        accept_none (str): whether accept None as the target value or not

    Raises:
        NAFoundError: @accept_none is False, but @target is None

    Note:
        When @accept_none is True and @target is None, default values will be returned with instance methods.
    """

    def __init__(self, target, name="target", accept_none=True):
        self._target = target
        self._name = str(name)
        if target is None and not accept_none:
            raise UnExpectedNoneError(self._name)

    def subclass(self, parent):
        """Ensure the target is a subclass of the parent class.

        Args:
            parent (object): parent class or sequence of parent classes

        Raises:
            NotSubclassError: the target is not the subclass

        Returns:
            object: the target itself
        """
        if issubclass(self._target, parent):
            return self._target
        raise NotSubclassError(self._name, self._target, parent)

    def instance(self, expected):
        """Ensure that the target is an instance of a specified class.

        Args:
            expected (object): expected class or sequence of expected classes

        Raises:
            UnExpectedTypeError: the target is not an instance of the class

        Returns:
            object: the target itself
        """
        if isinstance(self._target, expected):
            return self._target
        raise UnExpectedTypeError(self._name, self._target, expected)

    def dataframe(self, time_index=False, columns=None, empty_ok=True):
        """Ensure the target is a dataframe.

        Args:
            time_index (bool): if True, the dataframe must has DatetimeIndex
            columns (list[str] or None): the columns the dataframe must have
            empty_ok (bool): whether give permission to empty dataframe or not

        Raises:
            UnExpectedTypeError: the target is not a dataframe or that has un-expected-type index
            EmptyError: empty when @empty_ok is False
            NotIncludedError: expected columns were not included

        Returns:
            pandas.DataFrame: the target itself
        """
        if not isinstance(self._target, pd.DataFrame):
            raise UnExpectedTypeError(self._name, self._target, pd.DataFrame)
        df = self._target.copy()
        if not empty_ok and df.empty:
            raise EmptyError(name=self._name)
        if time_index and not isinstance(df.index, pd.DatetimeIndex):
            raise UnExpectedTypeError(f"Index of {self._name}", df.index, pd.DatetimeIndex)
        if columns is None:
            return df
        if not set(columns).issubset(df.columns):
            expected_cols = sorted(set(columns) - set(df.columns), key=columns.index)
            for col in expected_cols:
                raise NotIncludedError(
                    col, f"column list of {self._name}",
                    details=f"The dataframe has {', '.join(df.columns.tolist())} as columns")
        return df

    def float(self, value_range=(0, None), default=None, digits=None):
        """Convert a value to a float value.

        Args:
            value_range (tuple(int or None, int or None)): value range, None means un-specified
            default (float or None): default value when the target is None
            digits (int or None): effective digits or None (skip rounding)

        Raises:
            UnExpectedTypeError: the target cannot be converted to a float value
            UnExpectedValueRangeError: the value is out of value range

        Returns:
            float or None: converted float value or None (when both of the target and @default are None)
        """
        if self._target is None:
            return None if default is None else Validator(default, "default").float(value_range=value_range, digits=digits)
        try:
            value = float(self._target)
        except ValueError:
            raise UnExpectedTypeError(self._name, self._target, float) from None
        if (value < (value_range[0] or value)) or (value > (value_range[1] or value)):
            raise UnExpectedValueRangeError(self._name, value, value_range)
        if digits is None or value == 0:
            return value
        return round(value, digits - 1 - math.floor(math.log10(abs(value))))

    def int(self, value_range=(0, None), default=None, round_ok=False):
        """Convert a value to an integer.

        Args:
            value_range (tuple(int or None, int or None)): value range, None means un-specified
            default (int or None): default value when the target is None
            round_ok (bool): whether ignore round-off error

        Raises:
            UnExpectedTypeError: the target cannot be converted to an integer or round-off error exists when @round_ok is False
            UnExpectedValueRangeError: the value is out of value range

        Returns:
            int or None: converted float value or None (when both of the target and @default are None)
        """
        if self._target is None:
            return None if default is None else Validator(default, name="default").int(value_range=value_range, round_ok=round_ok)
        try:
            value = int(self._target)
        except (ValueError, TypeError):
            raise UnExpectedTypeError(self._name, self._target, int) from None
        if value != self._target and not round_ok:
            raise UnExpectedTypeError(
                self._name, self._target, int, details=f"This is because we cannot ignore round-off error, | {self._target} - {value} | > 0")
        if (value < (value_range[0] or value)) or (value > (value_range[1] or value)):
            raise UnExpectedValueRangeError(self._name, value, value_range)
        return value

    def tau(self, default=None):
        """Validate the value can be used as tau value [min].

        Args:
            default (int or None): default value when the target is None

        Raises:
            UnExpectedTypeError: the target cannot be converted to an integer
            UnExpectedValueRangeError: the value is out of value range

        Returns:
            int or None: converted float value or None (when both of the target and @default are None)
        """
        if self._target is None:
            return None if default is None else Validator(default, name="default").tau()
        value = self.int(value_range=(0, 1440), round_ok=False)
        if 1440 % value == 0:
            return value
        divisors = [str(i) for i in range(1, 1441) if 1440 % i == 0]
        raise UnExpectedValueError(
            self._name, value, divisors,
            details="Tau value [min], a divisor of 1440 [min] is a parameter used to convert actual time to time steps (without units)")

    def date(self, value_range=(None, None), default=None):
        """Convert a value to a date object.

        Args:
            value_range (tuple(int or None, int or None)): value range, None means un-specified
            default (pandas.Timestamp or None): default value when the target is None

        Raises:
            UnExpectedTypeError: the target cannot be converted to a date object
            UnExpectedValueRangeError: the value is out of value range

        Returns:
            pandas.Timestamp or None: converted date or None (when both of the target and @default are None)
        """
        if self._target is None:
            return None if default is None else Validator(default, name="default").date(value_range=value_range)
        if isinstance(self._target, pd.Timestamp):
            value = self._target.replace(hour=0, minute=0, second=0, microsecond=0)
        else:
            try:
                value = pd.to_datetime(self._target).replace(hour=0, minute=0, second=0, microsecond=0)
            except ValueError:
                raise UnExpectedTypeError(self._name, self._target, pd.Timestamp) from None
        if (value < (value_range[0] or value)) or (value > (value_range[1] or value)):
            raise UnExpectedValueRangeError(
                self._name, value.strftime("%Y-%m-%d"), [None if value is None else value.strftime("%Y-%m-%d") for value in value_range])
        return value

    def sequence(self, default=None, flatten=False, unique=False, candidates=None, length=None):
        """Convert a sequence (list, tuple) to a list.

        Args:
            default (list[object] or None): default value when the target is None
            flatten (bool): whether flatten the sequence or not
            unique (bool): whether remove duplicated values or not, the first value will remain
            candidates (list[object] or tuple(object) or iter or None): list of candidates or None (no limitations)
            length (int or None): length of the sequence or None (no limitations)

        Raises:
            UnExpectedTypeError: the target cannot be converted to a list or failed in flattening
            UnExpectedValueError: the target has a value which is not included in the candidates
            UnExpectedLengthError: the number of elements is not the same as @length

        Returns:
            list[object] or None: converted list or None (when both of the target and @default are None)
        """
        if self._target is None:
            return None if default is None else Validator(default, name="default").sequence(flatten=flatten, unique=unique, candidates=candidates)
        if not isinstance(self._target, (list, tuple)):
            raise UnExpectedTypeError(
                self._name, self._target, list, details="A tuple can be used, but it will be converted to a list")
        if flatten:
            try:
                targets = sum(self._target, [])
            except TypeError:
                for value in [value for value in self._target if not isinstance(value, list)]:
                    raise UnExpectedTypeError(
                        f"A value of {self._name}", value, list, details="This is required to flatten the sequence") from None
        else:
            targets = list(self._target)
        if unique:
            targets = sorted(set(targets), key=targets.index)
        if length is not None and len(targets) != length:
            raise UnExpectedLengthError(self._name, targets, length)
        if candidates is None or set(targets).issubset(candidates):
            return targets
        for value in (set(targets) - set(candidates)):
            raise UnExpectedValueError(self._name, value, [str(c) for c in candidates])

    def dict(self, default=None, required_keys=None, errors="coerce"):
        """Ensure the target is a dictionary.

        Args:
            default (dict[str, object] or None): default values, when the target is None or key is not included in the target
            required_keys (list): keys which must be included
            errors (str): "coerce" or "raise"

        Raises:
            UnExpectedTypeError: the target is not a dictionary
            NAFoundError: values of the required keys are not specified when @errors="raise"

        Returns:
            dict[str, object]: the target is self with default values and required keys

        Note:
            All keys of @default will be included and the target will overwrite it.

        Note:
            If some keys of @required_keys are not included and @errors="coerce", None will be set as the values of the keys.
        """
        if self._target is not None and not isinstance(self._target, dict):
            raise UnExpectedTypeError(self._name, self._target, dict)
        _dict = dict.fromkeys(required_keys or [])
        _dict.update(default or {})
        _dict.update(self._target or {})
        if required_keys is not None and errors != "coerce" and None in [_dict[key] for key in required_keys]:
            for key in [key for key in required_keys if _dict[key] is None]:
                raise NAFoundError(f"The value of key {key} in dictionary {self._name}")
        return _dict

    def kwargs(self, functions, default=None):
        """Find keyword arguments of the functions.

        Args:
            functions (list[function] or function): target functions
            default (dict[str, object] or None): default values, when the target is None or key is not included in the target

        Raises:
            UnExpectedTypeError: the target is not a dictionary

        Returns:
            dict[str, object]: keyword arguments of the functions
        """
        _dict = self.dict(default=default, required_keys=None, errors="coerce")
        keywords_nest = [
            list(signature(func).parameters.keys()) for func in (functions if isinstance(functions, list) else [functions])]
        keywords_set = set(sum(keywords_nest, [])) - {"self", "cls"}
        return {k: v for k, v in _dict.items() if k in keywords_set}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import warnings
import pandas as pd
from covsirphy.util.error import deprecate, DBLockedError, NotDBLockedError, UnExpectedValueError
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated.jhu_data import JHUData
from covsirphy.cleaning.japan_data import JapanData
from covsirphy._deprecated.oxcgrt import OxCGRTData
from covsirphy._deprecated.population import PopulationData
from covsirphy.cleaning.pyramid import PopulationPyramidData
from covsirphy._deprecated.linelist import LinelistData
from covsirphy._deprecated.pcr_data import PCRData
from covsirphy._deprecated.vaccine_data import VaccineData
from covsirphy._deprecated.mobility_data import MobilityData
from covsirphy._deprecated.db_covid19dh import _COVID19dh
from covsirphy._deprecated.db_google import _GoogleOpenData
from covsirphy._deprecated.recommended import _Recommended


class DataLoader(Term):
    """
    Load/download datasets and perform data cleaning.

    Args:
        directory (str or pathlib.Path): directory to save downloaded datasets
        update_interval (int or None): update interval of downloading dataset or None (avoid downloading)
        country (str or None): country name of datasets to downloaded when @update_interval is an integer
        basename_dict (dict[str, str] or None): basename of downloaded CSV files,
            "covid19dh": COVID-19 Data Hub (default: covid19dh.csv),
            "owid": Our World In Data (default: ourworldindata.csv),
            "google: COVID-19 Open Data by Google Cloud Platform (default: google_cloud_platform.csv),
            "wbdata": World Bank Open Data (default: wbdata_population_pyramid.csv),
            "japan": COVID-19 Dataset in Japan (default: covid_japan.csv).
        verbose (int): level of verbosity when downloading

    Note:
        If @update_interval (not None) hours have passed since the last update of downloaded datasets,
        the downloaded datasets will be updated automatically.
        When we do not use datasets of remote servers, set @update_interval as None.
        Note that the outputs of .japan() and .pyramid() are not locked by .lock().
        So, they are not influenced by @update_interval.

    Note:
        If @verbose is 0, no descriptions will be shown.
        If @verbose is 1 or larger, URL and database name will be shown.
        If @verbose is 2, detailed citation list will be show, if available.
    """

    @deprecate(old="DataLoader", new="DataEngineer", version="2.24.0-xi")
    def __init__(self, directory="input", update_interval=12, country=None, basename_dict=None, verbose=1):
        # Directory
        try:
            self.dir_path = Path(directory)
        except TypeError as e:
            raise TypeError(f"@directory should be a path-like object, but {directory} was applied.") from e
        self.dir_path.mkdir(parents=True, exist_ok=True)
        # Dictionary of filenames to save remote datasets
        file_dict = {
            "covid19dh": "covid19dh.csv",
            "owid": "ourworldindata.csv",
            "google": "google_cloud_platform.csv",
            "wbdata_pyramid": "wbdata_population_pyramid.csv",
            "japan": "covid_japan.csv",
        }
        file_dict.update(
            {k: self.dir_path.joinpath((basename_dict or {}).get(k, v)) for (k, v) in file_dict.items()}
        )
        self._file_dict = file_dict.copy()
        # Verbosity
        self._verbose = Validator(verbose, "verbose").int(value_range=(0, None))
        # Recommended datasets
        self._use_recommended = update_interval is not None
        self._recommended = None if update_interval is None else _Recommended(
            update_interval=update_interval, country=country, file_dict=self._file_dict, verbose=self._verbose)
        # Column names to identify records
        self._id_cols = [self.ISO3, self.PROVINCE, self.DATE]
        # Datasets retrieved from local files
        self._local_df = pd.DataFrame()
        self._local_citations = []
        # Flexible columns
        self._oxcgrt_cols = []
        self._mobility_cols = []
        # Locked database
        self._locked_df = pd.DataFrame(columns=self._id_cols)
        self._locked_citation_dict = {}

    @property
    def local(self):
        """
        pandas.DataFrame: local dataset
        """
        return self._local_df

    @property
    def locked(self):
        """
        pandas.DataFrame: locked dataset
        """
        self._ensure_lock_status(lock_expected=True)
        return self._locked_df

    def _ensure_lock_status(self, lock_expected):
        """
        Check whether the local database has been locked or not.

        Args:
            lock_expected (bool): whether the local database is expected to be locked or not

        Raises:
            NotDBLockedError: @lock_expected is True, but the database has NOT been locked
            DBLockedError: @lock_expected is False, but the database has been locked
        """
        if lock_expected and self._locked_df.empty:
            raise NotDBLockedError(name="the local database")
        if not lock_expected and not self._locked_df.empty:
            raise DBLockedError(name="the local database")

    def _edit_local(self, data, citation, how_combine, **kwargs):
        """
        Edit the local database with a dataframe.

        Args:
            data (pandas.DataFrame): dataframe to read
            citation (str or None): citation of the CSV file or None (basename of the CSV file)
            how_combine (str): how to combine datasets when we call this method multiple times
                - 'replace': replace registered dataset with the new data
                - 'concat': concat datasets with pandas.concat()
                - 'merge': merge datasets with pandas.DataFrame.merge()
                - 'update': update the current dataset with pandas.DataFrame.update()
            kwargs: keyword arguments of pandas.concat()/pandas.DataFrame.merge()/pandas.DataFrame.update()

        Raises:
            UnExpectedValueError: un-expected value was applied as @how_combine
        """
        v = Validator(kwargs, "keyword arguments")
        self._ensure_lock_status(lock_expected=False)
        if self._local_df.empty or how_combine == "replace":
            self._local_df = data.copy()
        elif how_combine == "concat":
            self._local_df = pd.concat(
                [self._local_df, data], ignore_index=True, sort=True, **v.kwargs(functions=pd.concat, default=None))
        elif how_combine == "merge":
            self._local_df = self._local_df.merge(data, **v.kwargs(functions=pd.merge, default=None))
        elif how_combine == "update":
            self._local_df.update(data, **v.kwargs(functions=data.update, default=None))
        else:
            raise UnExpectedValueError(
                "how_combine", how_combine, candidates=["replace", "concat", "merge", "update"])
        self._local_citations.append(citation)

    def read_csv(self, filename, citation=None, parse_dates=False, dayfirst=False,
                 how_combine="replace", **kwargs):
        """
        Read dataset saved in a CSV file and include it local database.

        Args:
            filename (str or pathlib.Path): path/URL of the CSV file
            citation (str or None): citation of the CSV file or None (basename of the CSV file)
            parse_dates (list[str] or bool): list of column names to parse as date information
            dayfirst (bool): whether date format is DD/MM or not
            how_combine (str): how to combine datasets when we call this method multiple times
                - 'replace': replace registered dataset with the new data
                - 'concat': concat datasets with pandas.concat()
                - 'merge': merge datasets with pandas.DataFrame.merge()
                - 'update': update the current dataset with pandas.DataFrame.update()
            kwargs: keyword arguments of pandas.read_csv()
                and pandas.concat()/pandas.DataFrame.merge()/pandas.DataFrame.update().

        Raises:
            UnExpectedValueError: un-expected value was applied as @how_combine

        Returns:
            covsirphy.DataLoader: self

        Note:
            Please refer to https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html
            for the keyword arguments of pandas.read_csv().

        Note:
            Please refer to https://pandas.pydata.org/docs/reference/api/pandas.concat.html
            for the keyword arguments of pandas.concat().
            Note that we always use 'ignore_index=True' and 'sort=True'.

        Note:
            Please refer to https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.merge.html
            for the keyword arguments of pandas.DataFrame.merge().

        Note:
            Please refer to https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.update.html
            for the keyword arguments of pandas.DataFrame.update().
        """
        v = Validator(kwargs, "keyword arguments")
        df = pd.read_csv(
            filename, parse_dates=parse_dates, dayfirst=dayfirst, **v.kwargs(functions=pd.read_csv, default=None))
        self._edit_local(
            data=df, citation=str(citation or Path(filename).name), how_combine=how_combine, **kwargs)
        return self

    def read_dataframe(self, dataframe, citation=None, parse_dates=False, dayfirst=False,
                       how_combine="replace", **kwargs):
        """
        Read a pandas.DataFrame and include it local database.

        Args:
            dataframe (pandas.DataFrame): dataframe to read
            citation (str or None): citation of the CSV file or None (basename of the CSV file)
            parse_dates (list[str] or bool): list of column names to parse as date information
            dayfirst (bool): whether date format is DD/MM or not
            how_combine (str): how to combine datasets when we call this method multiple times
                - 'replace': replace registered dataset with the new data
                - 'concat': concat datasets with pandas.concat()
                - 'merge': merge datasets with pandas.DataFrame.merge()
                - 'update': update the current dataset with pandas.DataFrame.update()
            kwargs: keyword arguments of pandas.to_datetime and
                pandas.concat()/pandas.DataFrame.merge()/pandas.DataFrame.update()

        Raises:
            UnExpectedValueError: un-expected value was applied as @how_combine

        Returns:
            covsirphy.DataLoader: self

        Note:
            Please refer to https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html
            for the keyword arguments of pandas.read_csv().

        Note:
            Please refer to https://pandas.pydata.org/docs/reference/api/pandas.concat.html
            for the keyword arguments of pandas.concat().
            Note that we always use 'ignore_index=True' and 'sort=True'.

        Note:
            Please refer to https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.merge.html
            for the keyword arguments of pandas.DataFrame.merge().

        Note:
            Please refer to https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.update.html
            for the keyword arguments of pandas.DataFrame.update().
        """
        df = Validator(dataframe, "dataframe").dataframe()
        if parse_dates:
            Validator(parse_dates, "parse_dates").sequence(candidates=df.columns)
            datetime_kwargs = Validator(kwargs, "keyword arguments").kwargs(functions=pd.to_datetime)
            for col in parse_dates:
                df[col] = pd.to_datetime(df[col], dayfirst=dayfirst, **datetime_kwargs)
        self._edit_local(
            data=dataframe, citation=str(citation or "dataframe"), how_combine=how_combine, **kwargs)
        return self

    def assign(self, **kwargs):
        """
        Assign new columns to the dataset retrieved from local files.

        Args:
            kwargs: keyword arguments of pandas.DataFrame.assign().

        Returns:
            covsirphy.DataLoader: self

        Note:
            Please refer to https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.assign.html
            for the keyword arguments of pandas.DataFrame.assign().
        """
        self._ensure_lock_status(lock_expected=False)
        self._local_df = self._local_df.assign(**kwargs)
        return self

    def lock(self, date, country=None, province=None, iso3=None,
             confirmed=None, fatal=None, recovered=None, population=None, tests=None,
             product=None, vaccinations=None, vaccinations_boosters=None, vaccinated_once=None, vaccinated_full=None,
             oxcgrt_variables=None, mobility_variables=None):
        """
        Lock the local database, specifying columns which has date and area information.

        Args:
            date (str): column name for dates
            country (str or None): country names (top level administration)
            province (str or None): province names (2nd level administration)
            iso3 (str or None): ISO3 codes
            confirmed (str or None): the number of confirmed cases
            fatal (str or None): the number of fatal cases
            recovered (str or None): the number of recovered cases
            population (str or None): population values
            tests (str or None): the number of tests
            product (str or None): vaccine product names
            vaccinations (str or None): cumulative number of vaccinations
            vaccinations_boosters (str or None): cumulative number of booster vaccinations
            vaccinated_once (str or None): cumulative number of people who received at least one vaccine dose
            vaccinated_full (str or None): cumulative number of people who received all doses prescribed by the protocol
            oxcgrt_variables (list[str] or None): list of variables for OxCGRTData
            mobility_variables (list[str] or None): list of variables for MobilityData

        Raises:
            ValueError: neither @country nor @iso3 is not specified

        Returns:
            covsirphy.DataLoader: self

        Note:
            If @oxcgrt_variables is None, variables registered in COVID-19 Data Hub will be used.

        Note:
            If @mobility_variables is None, variables registered in COVID-19 Open Data (Google) will be used.
        """
        self._ensure_lock_status(lock_expected=False)
        # Flexible variables
        self._oxcgrt_cols = oxcgrt_variables or _COVID19dh.OXCGRT_VARS[:]
        self._mobility_cols = mobility_variables or _GoogleOpenData.MOBILITY_VARS[:]
        # All variables
        variables = [
            self.COUNTRY, self.C, self.F, self.R, self.N, self.TESTS,
            self.PRODUCT, self.VAC, self.VAC_BOOSTERS, self.V_ONCE, self.V_FULL,
            *self._oxcgrt_cols, *self._mobility_cols,
        ]
        id_dict = {date: self.DATE, iso3: self.ISO3, province: self.PROVINCE}
        rename_dict = {
            **id_dict, country: self.COUNTRY,
            confirmed: self.C, fatal: self.F, recovered: self.R, population: self.N,
            tests: self.TESTS, product: self.PRODUCT, vaccinations: self.VAC, vaccinations_boosters: self.VAC_BOOSTERS,
            vaccinated_once: self.V_ONCE, vaccinated_full: self.V_FULL,
        }
        # Local database
        df = self._local_df.rename(columns=rename_dict)
        df = df.reindex(columns=[*self._id_cols, *variables])
        if country is not None and iso3 is None:
            df[self.ISO3] = self._to_iso3(df[self.COUNTRY])
        if df.empty:
            citation_dict = dict.fromkeys(variables, [])
        else:
            df[self.DATE] = pd.to_datetime(df[self.DATE])
            citation_dict = {v: self._local_citations if v in df else [] for v in variables}
            df = df.pivot_table(
                values=variables, index=self.DATE, columns=[self.ISO3, self.PROVINCE], aggfunc="first")
            df = df.resample("D").first().ffill().bfill()
            df = df.stack().stack().reset_index()
        # With Remote datasets
        if self._use_recommended:
            df = df.set_index(self._id_cols)
            remote_df, remote_dict = self._recommended.retrieve()
            df = df.combine_first(remote_df.set_index(self._id_cols)).reset_index()
            # Update citations
            citation_dict = {k: [*v, remote_dict[k]] if k in remote_df else v for (k, v) in citation_dict.items()}
        # Complete database lock
        all_cols = [*self._id_cols, *variables, *df.columns.tolist()]
        df = df.reindex(columns=sorted(set(all_cols), key=all_cols.index))
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        df[self.ISO3] = df[self.ISO3].fillna(self.NA)
        if self.COUNTRY in df:
            df[self.COUNTRY] = df[self.COUNTRY].fillna(self.NA)
        df[self.PROVINCE] = df[self.PROVINCE].fillna(self.NA)
        self._locked_df = df.dropna(subset=[self.DATE], axis=0)
        self._locked_citation_dict = citation_dict.copy()
        return self

    def _read_dep(self, basename=None, basename_owid=None, local_file=None, verbose=None):
        """
        Read deprecated keyword arguments.

        Args:
            basename (str or None): basename of the file to save the data
            basename_owid (str or None): basename of the file to save the data of "Our World In Data"
            local_file (str or None): if not None, load the data from this file
            verbose (int): level of verbosity
        """
        if basename is not None:
            raise TypeError("@basename argument was deprecated. Please use DataLoader(basename_dict)")
        if basename_owid is not None:
            raise TypeError("@basename_owid argument was deprecated. Please use DataLoader(basename_dict)")
        if local_file is not None:
            raise TypeError("local_file argument was deprecated. Please use DataLoader.read_csv().")
        if verbose is not None:
            warnings.warn(
                "verbose argument was deprecated. Please use DataLoader(verbose).",
                DeprecationWarning, stacklevel=2)
            self._verbose = Validator(verbose, "verbose").int(value_range=(0, None))

    def _auto_lock(self, variables):
        """
        Automatic database lock before using database.

        Args:
            variables (list[str] or None): variables to check citations

        Returns:
            tuple(pandas.DataFrame, dict[str, list[str]]):
                - locked database
                - citation list of the variables
        """
        # Database lock
        try:
            self._ensure_lock_status(lock_expected=True)
        except NotDBLockedError:
            self.lock(*self._id_cols)
        # Citation list
        if variables is None:
            return (self._locked_df, [])
        citation_dict = self._locked_citation_dict.copy()
        citations_nest = [c for v, line in citation_dict.items() for c in line if v in variables]
        citations = Validator(citations_nest).sequence(flatten=True, unique=True)
        return (self._locked_df, citations)

    @property
    def covid19dh_citation(self):
        """
        str: the list of primary sources of COVID-19 Data Hub.
        """
        self._auto_lock(variables=None)
        return self._recommended.covid19dh_citation()

    def jhu(self, **kwargs):
        """
        Load the dataset regarding the number of cases using local CSV file or COVID-19 Data Hub.

        Args:
            kwargs: all keyword arguments will be ignored

        Returns:
            covsirphy.JHUData: dataset regarding the number of cases
        """
        self._read_dep(**kwargs)
        df, citations = self._auto_lock(variables=[self.C, self.F, self.R, self.N])
        return JHUData(data=df, citation="\n".join(citations))

    @deprecate("DataLoader.population()", new="DataLoader.jhu()", version="2.21.0-xi-fu1")
    def population(self, **kwargs):
        """
        Deprecated, please use DataLoader.jhu() because JHUData includes population values.
        Load the dataset regarding population values using local CSV file or COVID-19 Data Hub.

        Args:
            kwargs: all keyword arguments will be ignored

        Returns:
            covsirphy.PopulationData: dataset regarding population values
        """
        self._read_dep(**kwargs)
        df, citations = self._auto_lock(variables=[self.N])
        return PopulationData(data=df, citation="\n".join(citations))

    def oxcgrt(self, **kwargs):
        """
        Load the dataset regarding OxCGRT indicators using local CSV file or COVID-19 Data Hub.

        Args:
            kwargs: all keyword arguments will be ignored

        Returns:
            covsirphy.JHUData: dataset regarding OxCGRT data
        """
        self._read_dep(**kwargs)
        df, citations = self._auto_lock(variables=[self._oxcgrt_cols])
        return OxCGRTData(data=df, citation="\n".join(citations), variables=self._oxcgrt_cols)

    def japan(self, **kwargs):
        """
        Load the dataset of the number of cases in Japan.
        https://github.com/lisphilar/covid19-sir/tree/master/data

        Args:
            kwargs: all keyword arguments will be ignored

        Returns:
            covsirphy.JapanData: dataset at country level in Japan
        """
        self._read_dep(**kwargs)
        filename = self._file_dict["japan"]
        return JapanData(filename=filename, force=True, verbose=self._verbose)

    @deprecate("DataLoader.linelist()", version="2.21.0-theta")
    def linelist(self, basename="linelist.csv", verbose=1):
        """
        Load linelist of case reports.
        https://github.com/beoutbreakprepared/nCoV2019

        Args:
            basename (str): basename of the file to save the data
            verbose (int): level of verbosity

        Returns:
            covsirphy.LinelistData: linelist data
        """
        filename = self.dir_path.joinpath(basename)
        return LinelistData(filename=filename, force=True, verbose=verbose)

    def pcr(self, **kwargs):
        """
        Load the dataset regarding the number of tests and confirmed cases,
        using local CSV file or COVID-19 Data Hub.

        Args:
            kwargs: all keyword arguments will be ignored

        Returns:
            covsirphy.PCRData: dataset regarding the number of tests and confirmed cases
        """
        self._read_dep(**kwargs)
        df, citations = self._auto_lock(variables=[self.TESTS, self.C])
        return PCRData(data=df, citation="\n".join(citations))

    def vaccine(self, **kwargs):
        """
        Load the dataset regarding vaccination.

        Args:
            kwargs: all keyword arguments will be ignored

        Returns:
            covsirphy.VaccineData: dataset regarding vaccinations
        """
        self._read_dep(**kwargs)
        v_cols = [self.PRODUCT, self.VAC, self.VAC_BOOSTERS, self.V_ONCE, self.V_FULL]
        df, citations = self._auto_lock(variables=v_cols)
        return VaccineData(data=df.dropna(subset=[self.VAC]), citation="\n".join(citations))

    def mobility(self):
        """
        Load the dataset regarding mobility.

        Returns:
            covsirphy.MobilityData: dataset regarding mobilities
        """
        df, citations = self._auto_lock(variables=self._mobility_cols)
        df = df.dropna(subset=self._mobility_cols)
        return MobilityData(data=df, citation="\n".join(citations), variables=self._mobility_cols)

    def pyramid(self, **kwargs):
        """
        Load the dataset regarding population pyramid.
        World Bank Group (2020), World Bank Open Data, https://data.worldbank.org/

        Args:
            kwargs: all keyword arguments will be ignored

        Returns:
            covsirphy.PopulationPyramidData: dataset regarding population pyramid
        """
        self._read_dep(**kwargs)
        filename = self._file_dict["wbdata_pyramid"]
        return PopulationPyramidData(filename=filename, force=False, verbose=self._verbose)

    def collect(self):
        """
        Collect data for scenario analysis and return them as a dictionary.

        Returns:
            dict(str, object):
                - jhu_data (covsirphy.JHUData)
                - extras (list[covsirphy.CleaningBase]):
                    - covsirphy.OXCGRTData
                    - covsirphy.PCRData
                    - covsirphy.VaccineData
                    - covsirphy.MobilityData
        """
        return {
            "jhu_data": self.jhu(),
            "extras": [self.oxcgrt(), self.pcr(), self.vaccine(), self.mobility()]
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from covsirphy.util.error import deprecate
from covsirphy._deprecated._mbase import ModelBase


class SIRFV(ModelBase):
    """
    SIR-FV model.

    Args:
        population (int): total population
            theta (float)
            kappa (float)
            rho (float)
            sigma (float)
            omega (float) or v_per_day (int)
    """
    # Model name
    NAME = "SIR-FV"
    # names of parameters
    PARAMETERS = ["theta", "kappa", "rho", "sigma", "omega"]
    DAY_PARAMETERS = [
        "alpha1 [-]", "1/alpha2 [day]", "1/beta [day]", "1/gamma [day]",
        "Vaccinated [persons]"
    ]
    # Variable names in (non-dim, dimensional) ODEs
    VAR_DICT = {
        "x": ModelBase.S,
        "y": ModelBase.CI,
        "z": ModelBase.R,
        "w": ModelBase.F,
        "v": ModelBase.V
    }
    VARIABLES = list(VAR_DICT.values())
    # Weights of variables in parameter estimation error function
    WEIGHTS = np.array([0, 10, 10, 2, 0])
    # Variables that increases monotonically
    VARS_INCREASE = [ModelBase.R, ModelBase.F]
    # Example set of parameters and initial values
    EXAMPLE = {
        ModelBase.STEP_N: 180,
        ModelBase.N.lower(): 1_000_000,
        ModelBase.PARAM_DICT: {
            "theta": 0.002, "kappa": 0.005, "rho": 0.2, "sigma": 0.075,
            "omega": 0.001,
        },
        ModelBase.Y0_DICT: {
            ModelBase.S: 999_000, ModelBase.CI: 1000, ModelBase.R: 0, ModelBase.F: 0,
            ModelBase.V: 0,
        },
    }

    @deprecate(old="ModelBase", new="ODEModel", version="2.24.0-xi")
    def __init__(self, *args, **kwargs):
        raise NotImplementedError(
            "SIR-FV model was removed because vaccinated persons may move "
            "to the other compartments. Please use SIR-F model and adjust parameter values "
            "of SIR-F model, considering the impact of vaccinations on infectivity, effectivity and safety."
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from covsirphy.util.term import Term
from covsirphy._deprecated.db_base import _RemoteDatabase


class _CSJapan(_RemoteDatabase):
    """
    Access "COVID-19 Dataset in Japan.
    https://github.com/lisphilar/covid19-sir/tree/master/data

    Args:
        filename (str): CSV filename to save records
        iso3 (str or None): ignored
    """
    # URL
    GITHUB_URL = "https://raw.githubusercontent.com"
    URL_C = f"{GITHUB_URL}/lisphilar/covid19-sir/master/data/japan/covid_jpn_total.csv"
    URL_P = f"{GITHUB_URL}/lisphilar/covid19-sir/master/data/japan/covid_jpn_prefecture.csv"
    # Citation
    CITATION = "Hirokazu Takaya (2020-2022), COVID-19 dataset in Japan, GitHub repository, " \
        "https://github.com/lisphilar/covid19-sir/data/japan"
    # Column names and data types
    # {"name in database": "name defined in Term class"}
    COL_DICT = {
        "Date": Term.DATE,
        Term.COUNTRY: Term.COUNTRY,
        "Area": Term.PROVINCE,
        Term.ISO3: Term.ISO3,
        "Positive": Term.C,
        "Fatal": Term.F,
        "Discharged": Term.R,
        "Hosp_require": "Hosp_require",
        Term.MODERATE: Term.MODERATE,
        "Hosp_severe": Term.SEVERE,
        "Tested": Term.TESTS,
        Term.VAC: Term.VAC,
        Term.VAC_BOOSTERS: Term.VAC_BOOSTERS,
        Term.V_ONCE: Term.V_ONCE,
        Term.V_FULL: Term.V_FULL,
    }

    def download(self, verbose):
        """
        Download the dataset from the server and set the list of primary sources.

        Args:
            verbose (int): level of verbosity

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    defined by the first values of self.COL_DICT.values()

        Note:
            If @verbose is equal to or over 1, how to show the list will be explained.
        """

        # Download datasets
        if verbose:
            print("Retrieving COVID-19 dataset from https://github.com/lisphilar/covid19-sir/data/")
        # Domestic/Airport/Returnee
        dar_value_cols = ["Positive", "Tested", "Discharged", "Fatal", "Hosp_require", "Hosp_severe"]
        dar_cols = [*dar_value_cols, "Date", "Location", "Vaccinated_1st", "Vaccinated_2nd", "Vaccinated_3rd"]
        dar_df = pd.read_csv(self.URL_C, usecols=dar_cols)
        dar_df = dar_df.rename(columns={"Location": "Area"}).set_index("Date")
        # Country level data
        c_df = dar_df.groupby("Date").sum().reset_index()
        c_df["Area"] = self.NA
        # Entering (= Airport + Returnee)
        e_df = dar_df.loc[dar_df["Area"].isin(["Airport", "Returnee"])].groupby("Date").sum().reset_index()
        e_df["Area"] = "Entering"
        # Province level data
        p_cols = [*dar_value_cols, "Date", "Prefecture"]
        p_df = pd.read_csv(self.URL_P, usecols=p_cols)
        p_df = p_df.rename(columns={"Prefecture": "Area"})
        # Combine
        df = pd.concat([c_df, e_df, p_df], axis=0, ignore_index=True, sort=True)
        # Set additional columns
        df[self.COUNTRY] = "Japan"
        df[self.ISO3] = "JPN"
        df[self.MODERATE] = df["Hosp_require"] - df["Hosp_severe"]
        df[self.V_ONCE] = df["Vaccinated_1st"].cumsum()
        df[self.V_FULL] = df["Vaccinated_2nd"].cumsum()
        df[self.VAC_BOOSTERS] = df["Vaccinated_3rd"].cumsum()
        df[self.VAC] = df[[self.V_ONCE, self.V_FULL, self.VAC_BOOSTERS]].sum(axis=1)
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from covsirphy.util.error import deprecate, SubsetNotFoundError
from covsirphy.cleaning.cbase import CleaningBase


class PopulationData(CleaningBase):
    """
    Deprecated.
    Data cleaning of total population dataset.

    Args:
        filename (str or None): CSV filename of the dataset
        data (pandas.DataFrame or None):
            Index
                reset index
            Columns
                - Date: Observation date
                - ISO3: ISO3 code
                - Country: country/region name
                - Province: province/prefecture/state name
                - Population: total population
        citation (str or None): citation or None (empty)

    Note:
        Either @filename (high priority) or @data must be specified.
    """

    @deprecate("PopulationData", new="JHUData", version="2.21.0-xi-fu1")
    def __init__(self, filename=None, data=None, citation=None):
        super().__init__(filename=filename, data=data, citation=citation, variables=[self.N])

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - ISO3 (str): ISO3 code or "-"
                    - Country (pandas.Category): country/region name
                    - Province (pandas.Category): province/prefecture/state name
                    - Date (pd.Timestamp): date of the records (if available) or today
                    - Population (int): total population
        """
        df = self._raw.copy()
        # Date
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        # Province
        df[self.PROVINCE] = df[self.PROVINCE].fillna(self.NA)
        df[self.N] = df[self.N].fillna(0).astype(np.int64)
        # Columns to use
        df = df.loc[:, self._raw_cols]
        # Remove duplicates
        df = df.dropna(subset=[self.N]).drop_duplicates().reset_index(drop=True)
        # Update data types to reduce memory
        df[self.AREA_ABBR_COLS] = df[self.AREA_ABBR_COLS].astype("category")
        return df

    def total(self):
        """
        Return the total value of population in the dataset.

        Returns:
            int
        """
        values = self._cleaned_df[self.N]
        return int(sum(values))

    def to_dict(self, country_level=True):
        """
        Return dictionary of population values.

        Args:
        country_level (str): whether key is country name or not

        Returns:
            dict
                - if @country_level is True, {"country", population}
                - if False, {"country/province", population}
        """
        df = self._cleaned_df.copy()
        if country_level:
            df = df.loc[df[self.PROVINCE] == self.NA, :]
            df["key"] = df[self.COUNTRY]
        else:
            df = df.loc[df[self.PROVINCE] != self.NA, :]
            df["key"] = df[self.COUNTRY].str.cat(df[self.PROVINCE], sep=self.SEP)
        return df.set_index("key").to_dict()[self.N]

    def value(self, country, province=None, date=None):
        """
        Return the value of population in the place.

        Args:
            country (str): country name or ISO3 code
            province (str): province name
            date (str or None): observation date, like 01Jun2020

        Returns:
            int: population in the place

        Note:
            If @date is None, the created date of the instancewill be used
        """
        country_alias = self.ensure_country_name(country)
        try:
            df = self.subset(country=country, province=province, start_date=date, end_date=date)
        except KeyError:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province, date=date)
        df = df.sort_values(self.DATE)
        return int(df.loc[df.index[-1], [self.N]].values[0])

    def update(self, value, country, province=None, date=None):
        """
        Update the value of a new place.

        Args:
            value (int): population in the place
            country (str): country name
            province (str): province name
            date (str or None): observation date, like 01Jun2020

        Returns:
            covsirphy.PopulationData: self

        Note:
            If @date is None, the created date of the instance will be used.
            If @province is None, "-" will be used.
        """
        population = self._ensure_natural_int(value, "value")
        province = province or self.NA
        date_stamp = self._ensure_date(date, name="date", default=self._cleaned_df[self.DATE].max())
        df = self._cleaned_df.copy()
        c_series = df[self.COUNTRY]
        p_series = df[self.PROVINCE]
        d_series = df[self.DATE]
        sel = (c_series == country) & (p_series == province) & (d_series == date_stamp)
        if not df.loc[sel, :].empty:
            df.loc[sel, self.N] = value
            self._cleaned_df = df.copy()
            return self
        series = pd.Series(
            [date_stamp, self.NA, country, province, population], index=self._raw_cols)
        self._cleaned_df = df.append(series, ignore_index=True)
        return self

    def countries(self):
        """
        Return names of countries where records are registered.

        Raises:
            KeyError: Country names are not registered in this dataset

        Returns:
            list[str]: list of country names

        Note:
            Country 'Others' will be removed.
        """
        country_list = super().countries()
        removed_countries = ["Others"]
        country_list = list(set(country_list) - set(removed_countries))
        return country_list

    def map(self, country=None, variable="Population", date=None, **kwargs):
        """
        Create colored map with the number of tests.

        Args:
            country (str or None): country name or None (global map)
            variable (str): always 'Population'
            date (str or None): date of the records or None (the last value)
            kwargs: arguments of ColoredMap() and ColoredMap.plot()

        Raises:
            NotImplementedError: @variable was specified

        Note:
            When @country is None, country level data will be shown on global map.
            When @country is a country name, province level data will be shown on country map.
        """
        if variable != self.N:
            raise NotImplementedError(f"@variable cannot be changed, always {self.N}.")
        # Date
        date_str = date or self.cleaned()[self.DATE].max().strftime(self.DATE_FORMAT)
        country_str = country or "Global"
        title = f"{country_str}: {variable.lower()} on {date_str}"
        # Global map
        if country is None:
            return self._colored_map_global(variable=variable, title=title, date=date, **kwargs)
        # Country-specific map
        return self._colored_map_country(
            country=country, variable=variable, title=title, date=date, **kwargs)


class Population(PopulationData):
    @deprecate(old="Population()", new="PopulationData()")
    def __init__(self, filename):
        super().__init__(filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import matplotlib
if not hasattr(sys, "ps1"):
    matplotlib.use("Agg")
from matplotlib import pyplot as plt
import seaborn as sns
from covsirphy.visualization.vbase import VisualizeBase


class _PredActualPlot(VisualizeBase):
    """
    Class for a scatter plot (predicted vs. actual parameter values).

    Args:
        filename (str or None): filename to save the figure or None (display)
        bbox_inches (str): bounding box in inches when creating the figure
        kwargs: the other arguments of matplotlib.pyplot.savefig
    """

    def __init__(self, filename=None, bbox_inches="tight", **kwargs):
        self._filename = filename
        self._savefig_dict = {"bbox_inches": bbox_inches, **kwargs}
        # Properties
        self._title = ""
        self._ax = None
        self._facet_grid = None

    def __enter__(self):
        return super().__enter__()

    def __exit__(self, *exc_info):
        # Settings
        if self._title:
            self._facet_grid.fig.subplots_adjust(top=0.9)
            self._facet_grid.fig.suptitle(self._title)
        # Display the figure if filename is None after plotting
        if self._filename is None:
            plt.show()
        else:
            # Save the image as a file
            self._facet_grid.savefig(self._filename, **self._savefig_dict)
            plt.clf()
            plt.close("all")

    def plot(self, data, x, y):
        """
        Create a scatter plot (predicted vs. actual parameter values).

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - {x} (float): actual values
                    - {y} (float): predicted values
                    - parameter (float): parameter names
                    - subset (str): "Train" or "Test"
            x (str): x label
            y (str): y label
        """
        facet_grid = sns.relplot(
            data=data, x=x, y=y, col="parameter", hue="subset", style="subset", kind="scatter",
            col_wrap=2, alpha=0.7, facet_kws={"sharex": False, "sharey": False, "legend_out": False})
        # Add y = x lines
        for ax in facet_grid.axes.flat:
            ax.axline((0, 0), slope=1, ls=":")
        # Set limits
        facet_grid.set(xlim=(0, None), ylim=(0, None))
        # Set Legend
        handles, legends = facet_grid.axes[0].get_legend_handles_labels()
        facet_grid.axes[0].legend_.remove()
        facet_grid.fig.legend(
            handles, legends, title="", ncol=2, loc="lower center", bbox_to_anchor=(0.5, -0.06))
        # Save grid
        self._facet_grid = facet_grid



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.util.error import deprecate
from covsirphy.util.term import Term


class PhaseUnit(Term):
    """
    Save information of  a phase.

    Args:
        start_date (str): start date of the phase
        end_date (str): end date of the phase
        population (int): population value

    Examples:
        >>> unit1 = PhaseUnit("01Jan2020", "01Feb2020", 1000)
        >>> unit2 = PhaseUnit("02Feb2020", "01Mar2020", 1000)
        >>> unit3 = PhaseUnit("02Mar2020", "01Apr2020", 1000)
        >>> unit4 = PhaseUnit("02Mar2020", "01Apr2020", 1000)
        >>> unit5 = PhaseUnit("01Jan2020", "01Apr2020", 1000)
        >>> str(unit1)
        'Phase (01Jan2020 - 01Feb2020)'
        >>> unit4 == unit4
        True
        >>> unit1 != unit2
        True
        >>> unit1 < unit2
        True
        >>> unit3 > unit1
        True
        >>> unit3 < unit4
        False
        >>> unit3 <= unit4
        True
        >>> unit1 < "02Feb2020"
        True
        >>> unit1 <= "01Feb2020"
        True
        >>> unit1 > "31Dec2019"
        True
        >>> unit1 >= "01Jan2020"
        True
        >>> sorted([unit3, unit1, unit2]) == [unit1, unit2, unit3]
        True
        >>> str(unit1 + unit2)
        'Phase (01Jan2020 - 01Mar2020)'
        >>> str(unit5 - unit1)
        'Phase (02Feb2020 - 01Apr2020)'
        >>> str(unit5 - unit4)
        'Phase (01Jan2020 - 01Mar2020)'
        >>> set([unit1, unit3, unit4]) == set([unit1, unit3])
        True
    """

    @deprecate("PhaseUnit", new="ODEHandler", version="2.19.1-zeta-fu1")
    def __init__(self, start_date, end_date, population):
        raise NotImplementedError

    def __str__(self):
        if self._id_dict is None:
            header = "Phase"
        else:
            id_str = ', '.join(list(self._id_dict.values()))
            header = f"{id_str:>4} phase"
        return f"{header} ({self._start_date} - {self._end_date})"

    def __hash__(self):
        return hash((self._start_date, self._end_date))

    def __eq__(self, other):
        if not isinstance(other, PhaseUnit):
            raise NotImplementedError
        return self._start_date == other.start_date and self._end_date == other.end_date

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        # self < other
        end = self._ensure_date(self._end_date)
        if isinstance(other, str):
            sta_other = self._ensure_date(other)
        elif isinstance(other, PhaseUnit):
            sta_other = self._ensure_date(other.start_date)
        else:
            raise NotImplementedError
        return end < sta_other

    def __le__(self, other):
        # self <= other
        end = self._ensure_date(self._end_date)
        if isinstance(other, str):
            sta_other = self._ensure_date(other)
        elif isinstance(other, PhaseUnit):
            if self.__eq__(other):
                return True
            sta_other = self._ensure_date(other.start_date)
        else:
            raise NotImplementedError
        return end <= sta_other

    def __gt__(self, other):
        # self > other
        if isinstance(other, PhaseUnit) and self.__eq__(other):
            return False
        return not self.__le__(other)

    def __ge__(self, other):
        # self >= other
        return not self.__lt__(other)

    def __add__(self, other):
        if self < other:
            return PhaseUnit(self._start_date, other.end_date, self._population)
        raise NotImplementedError

    def __iadd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        sta = self._ensure_date(self._start_date)
        end = self._ensure_date(self._end_date)
        sta_other = self._ensure_date(other.start_date)
        end_other = self._ensure_date(other.end_date)
        if sta < sta_other and end == end_other:
            end_date = self.yesterday(other.start_date)
            return PhaseUnit(self._start_date, end_date, self._population)
        if sta == sta_other and end > end_other:
            start_date = self.tomorrow(other.end_date)
            return PhaseUnit(start_date, self._end_date, self._population)

    def __isub__(self, other):
        return self.__sub__(other)

    def __contains__(self, date):
        sta = self._ensure_date(self._start_date)
        end = self._ensure_date(self._end_date)
        date = self._ensure_date(date)
        return sta <= date <= end



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import timedelta
import numpy as np
import pandas as pd
from covsirphy.util.error import deprecate, NAFoundError
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class ModelBase(Term):
    """
    Base class of ODE models.
    """
    # Model name
    NAME = "ModelBase"
    # names of parameters
    PARAMETERS = list()
    DAY_PARAMETERS = list()
    # Variable names in (non-dim, dimensional) ODEs
    VAR_DICT = dict()
    VARIABLES = list(VAR_DICT.values())
    # Weights of variables in parameter estimation error function
    WEIGHTS = np.array(list())
    # Variables that increases monotonically
    VARS_INCREASE = list()
    # Example set of parameters and initial values
    EXAMPLE = {
        Term.STEP_N: 180,
        Term.N.lower(): 1_000_000,
        Term.PARAM_DICT: dict(),
        Term.Y0_DICT: dict(),
    }

    @deprecate(old="ModelBase", new="ODEModel", version="2.24.0-xi")
    def __init__(self, population):
        """
        This method should be overwritten in subclass.

        Args:
            population (int): total population
        """
        # Total population
        self.population = Validator(population, "population").int(value_range=(1, None))
        # Dictionary of non-dim parameters: {name: value}
        self.non_param_dict = {}

    def __str__(self):
        param_str = ", ".join(f"{p}={v}" for (
            p, v) in self.non_param_dict.items())
        return f"{self.NAME} model with {param_str}"

    def __getitem__(self, key):
        """
        Args:
            key (str): parameter name
        """
        if key not in self.non_param_dict.keys():
            raise KeyError(f"key must be in {', '.join(self.PARAMETERS)}")
        return self.non_param_dict[key]

    def __call__(self, t, X):
        """
        Return the list of dS/dt (tau-free) etc.
        This method should be overwritten in subclass.

        Returns:
            np.array
        """
        raise NotImplementedError

    @classmethod
    @deprecate(".param_range()", new=".guess()", version="2.19.1-zeta-fu1")
    def param_range(cls, taufree_df, population, quantiles=(0.3, 0.7)):
        """
        Deprecated. Define the value range of ODE parameters using (X, dX/dt) points.
        This method should be overwritten in subclass.

        Args:
            taufree_df (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - t (int): time steps (tau-free)
                    - columns with dimensional variables
            population (int): total population
            quantiles (tuple(int, int)): quantiles to cut, like confidence interval

        Returns:
            dict(str, tuple(float, float)): minimum/maximum values
        """
        raise NotImplementedError

    @classmethod
    @deprecate(".specialize()", new=".convert()", version="2.19.1-zeta-fu1")
    def specialize(cls, data_df, population):
        """
        Deprecated. Specialize the dataset for this model.
        This method should be overwritten in subclass.

        Args:
            data_df (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - any columns
            population (int): total population in the place

        Returns:
            (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - any columns @data_df has
                    - columns with dimensional variables
        """
        raise NotImplementedError

    @classmethod
    @deprecate(".restore()", new=".convert_reverse()", version="2.19.1-zeta-fu1")
    def restore(cls, specialized_df):
        """
        Deprecated. Restore Confirmed/Infected/Recovered/Fatal using a dataframe with the variables of the model.
        This method should be overwritten in subclass.

        Args:
            specialized_df (pandas.DataFrame): dataframe with the variables

                Index
                    (object):
                Columns
                    - variables of the models (int)
                    - any columns

        Returns:
            (pandas.DataFrame):
                Index
                    (object): as-is
                Columns
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - the other columns @specialized_df has
        """
        raise NotImplementedError

    def calc_r0(self):
        """
        Calculate (basic) reproduction number.
        This method should be overwritten in subclass.

        Returns:
            float
        """
        raise NotImplementedError

    def calc_days_dict(self, tau):
        """
        Calculate 1/beta [day] etc.
        This method should be overwritten in subclass.

        Args:
            param tau (int): tau value [min]

        Returns:
            dict[str, int]
        """
        raise NotImplementedError

    @classmethod
    @deprecate(".taufree()", new=".convert()", version="2.19.1-zeta-fu1")
    def tau_free(cls, subset_df, population, tau=None):
        """
        Deprecated. Create a dataframe specialized to the model.
        If tau is not None, Date column will be converted to '(Date - start date) / tau'
        and saved in t column.

        Args:
            subset_df (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - any columns
            population (int): population value
            tau (int or None): tau value [min], 0 <= tau <= 1440

        Returns:
            (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - t (int): if tau is not None
                    - columns with dimensional variables
        """
        raise NotImplementedError

    @classmethod
    def _convert(cls, data, tau):
        """
        Divide dates by tau value [min].

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            tau (int): tau value [min] or None (skip division by tau values)

        Returns:
            pandas.DataFrame:
                Index
                    - Date (pd.Timestamp): Observation date (available when @tau is None)
                    - t (int): time steps (available when @tau is not None)
                Columns
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
        """
        df = Validator(data, "data").dataframe(columns=cls.DSIFR_COLUMNS)
        if tau is None:
            return df.set_index(cls.DATE)
        # Convert to tau-free
        tau = Validator(tau, "tau").tau(default=None)
        if tau is None:
            raise NAFoundError("tau", None)
        time_series = (pd.to_datetime(df[cls.DATE]) - df[cls.DATE].min()).dt.total_seconds() // 60
        df.index = (time_series / tau).astype(np.int64)
        df.index.name = cls.TS
        return df.drop(cls.DATE, axis=1)

    @classmethod
    def _convert_reverse(cls, converted_df, start, tau):
        """
        Calculate date with tau and start date.

        Args:
            converted_df (pandas.DataFrame):
                Index
                    t: Dates divided by tau value (time steps)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal or Recovered (int): the number of fatal/recovered cases
            start (pd.Timestamp): start date of simulation, like 14Apr2021
            tau (int): tau value [min]

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
        """
        df = converted_df.copy()
        # Calculate date with tau and start date
        tau = Validator(tau, "tau").tau(default=None)
        if tau is None:
            raise NAFoundError("tau", None)
        elapsed = pd.Series(df.index * tau)
        df[cls.DATE] = start + elapsed.apply(lambda x: timedelta(minutes=x))
        # Select the last records for dates
        df = df.set_index(cls.DATE).resample("D").last().reset_index()
        return df

    @classmethod
    def convert(cls, data, tau):
        """
        Divide dates by tau value [min] and convert variables to model-specialized variables.
        This will be overwritten by child classes.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            tau (int): tau value [min] or None (skip division by tau values)

        Returns:
            pandas.DataFrame:
                Index
                    time steps: Dates divided by tau value
                Columns
                    - model-specialized variables
        """
        raise NotImplementedError

    @classmethod
    def convert_reverse(cls, converted_df, start, tau):
        """
        Calculate date with tau and start date, and restore Susceptible/Infected/"Fatal or Recovered".
        This will be overwritten by child classes.

        Args:
            converted_df (pandas.DataFrame):
                Index
                    time steps: Dates divided by tau value
                Columns
                    - model-specialized variables
            start (pd.Timestamp): start date of simulation, like 14Apr2021
            tau (int): tau value [min]

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
        """
        raise NotImplementedError

    @classmethod
    def guess(cls, data, tau, q=0.5):
        """
        With (X, dX/dt) for X=S, I, R and so on, guess parameter values.
        This will be overwritten by child classes.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            tau (int): tau value [min]
            q (float or tuple(float,)): the quantile(s) to compute, value(s) between (0, 1)

        Returns:
            dict(str, float or pandas.Series): guessed parameter values with the quantile(s)
        """
        raise NotImplementedError

    @classmethod
    def _clip(cls, values, lower, upper):
        """
        Trim values at input threshold.

        Args:
            values (float or pandas.Series): values to trim
            lower (float): minimum threshold
            upper (float): maximum threshold
        """
        if isinstance(values, float):
            return min(max(values, lower), upper)
        Validator(values, "values").instance(pd.Series)
        return values.clip()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime, timedelta
from itertools import product
import pandas as pd
from covsirphy.util.error import deprecate
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated._mbase import ModelBase
from covsirphy.dynamics.ode import ODEModel
from covsirphy._deprecated.autots_predictor import _AutoTSPredictor


class AutoMLHandler(Term):
    """
    Predict ODE parameter values automatically with machine learning.

    Args:
        X (pandas.DataFrame):
            Index
                pandas.Timestamp: Observation date
            Columns
                observed variables (int or float)
        Y (pandas.DataFrame):
            Index
                pandas.Timestamp: Observation date
            Columns
                observed ODE parameter values (float)
        model (covsirphy.ModelBase or covsirphy.ODEModel): ODE model
        days (int): days to predict
        kwargs: keyword arguments of autots.AutoTS()

    Note:
        When X is a empty dataframe, only "univariate" can be used as @method with AutoHandler.predict().
    """
    _LIKELY = "Likely"
    _UPPER = "Upper"
    _LOWER = "Lower"

    @deprecate(old="AutoMLHandler()", new="ODEScenario().predict()", version="2.24.0-sigma")
    def __init__(self, X, Y, model, days, **kwargs):
        self._model = Validator(model, "model").subclass(parent=(ModelBase, ODEModel))
        if issubclass(self._model, ModelBase):
            self._parameters = self._model.PARAMETERS[:]
        else:
            self._parameters = self._model._PARAMETERS[:]
        self._X = Validator(X, "X").dataframe(time_index=True, empty_ok=True)
        self._Y = Validator(Y, "Y").dataframe(time_index=True, empty_ok=False, columns=self._parameters)
        self._days = Validator(days, name="days").int(value_range=(1, None))
        self._kwargs = kwargs.copy()
        self._pred_df = pd.DataFrame(columns=[self.SERIES, self.DATE, *Y.columns.tolist()])

    def predict(self, method):
        """
        Perform automated machine learning to predict values.

        Args:
            method (str): machine learning method name, "univariate" or "multivariate_regression"

        Returns:
            AutoMLHandler: self

        Note:
            Models used by "univariate" can be checked with from autots.models.model_list import model_lists; model_list["univariate"].

        Note:
            Model used by "multivariate_regression" is Multivariate Regression.
        """
        method_dict = {
            "univariate": self._univariate,
            "multivariate_regression": self._multivariate_regression,
        }
        if method not in method_dict:
            raise KeyError(
                f"Un-expected method: {method}. Supported methods are {', '.join(list(method_dict.keys()))}.")
        self._register_scenarios(method, *method_dict[method]())
        return self

    def summary(self):
        """
        Create and summarize the scenarios.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Scenario (str): scenario name, "Univariate_Likely", "Univariate_1" etc.
                    - Start (pandas.Timestamp): start date of the phase
                    - End (pandas.Timestamp): end date of the phase
                    - Rt (float): phase-dependent reproduction number
                    - columns of Y data

        Note:
            "Univariate_Likely" scenario is the most likely scenario when univariate forecasting is used.

        Note:
            "Univariate_01" scenario is the created with upper values of ODE parameter values.

        Note:
            "Univariate_16" scenario is the created with lower values of ODE parameter values. (if the model has four parameters)

        Note:
            Dates with the same Rt values at the 1st decimal place will be merged to one phase.
        """
        df = self._pred_df.copy()
        # Calculate reproduction number to create phases
        df["param"] = df[self._parameters].to_dict(orient="records")
        if issubclass(self._model, ModelBase):
            df[self.RT] = df.apply(lambda x: self._model(population=100, **x["param"]).calc_r0(), axis=1).round(1)
        else:
            model = self._model
            today = datetime.now()
            model_dict = {
                "date_range": (today, today + timedelta(days=10)),
                "tau": 1440,
                "initial_dict": model._SAMPLE_DICT["initial_dict"]
            }
            df[self.RT] = df[self._parameters].apply(lambda x: model(
                param_dict=x.to_dict(), **model_dict).r0(), axis=1).round(1)
        # Get start/end date
        criteria = [self.SERIES, self.RT]
        df = df.groupby(criteria).first().join(df[[*criteria, self.DATE]].groupby(criteria).last(), rsuffix="_last")
        df = df.rename(columns={self.DATE: self.START, f"{self.DATE}_last": self.END})
        df = df.reset_index().loc[:, [self.SERIES, self.START, self.END, self.RT, *self._parameters]]
        return df.sort_values([self.SERIES, self.START], ignore_index=True)

    def _register_scenarios(self, method, likely_df, upper_df, lower_df):
        """
        Create and register scenario with the most likely values, upper values and lower values.

        Args:
            method (str): machine learning method name
            likely_df (pandas.DataFrame): the most likely values with a forecasting method
                Index
                    Date (pandas.Timestamp): observation date
                Columns
                    predicted values (float)
            upper_df (pandas.DataFrame): the upper values with a forecasting method
                Index
                    Date (pandas.Timestamp): observation date
                Columns
                    predicted values (float)
            lower_df (pandas.DataFrame): the lower values with a forecasting method
                Index
                    Date (pandas.Timestamp): observation date
                Columns
                    predicted values (float)
        """
        # The most likely scenario
        df = likely_df.loc[:, self._Y.columns]
        df.index.name = self.DATE
        df = df.reset_index()
        df[self.SERIES] = f"{method.capitalize()}_{self._LIKELY}"
        dataframes = [df]
        # Upper/Lower
        ul_df = upper_df.loc[:, self._Y.columns].join(
            lower_df.loc[:, self._Y.columns], lsuffix=f"_{self._UPPER}", rsuffix=f"_{self._LOWER}")
        col_products = product(
            *([f"{param}_{suffix}" for suffix in (self._UPPER, self._LOWER)] for param in self._Y.columns))
        for (i, col_product) in enumerate(col_products):
            df = ul_df.loc[:, col_product]
            df.rename(
                columns=lambda x: x.replace(self._UPPER, "").replace(self._LOWER, "").replace("_", ""), inplace=True)
            df[self.SERIES] = f"{method.capitalize()}_{i:02}"
            df.index.name = self.DATE
            dataframes.append(df.reset_index())
        self._pred_df = pd.concat([self._pred_df, *dataframes], axis=0)

    def _univariate(self):
        """
        Perform univariate forecasting of Y without X.

        Returns:
            tuple(pandas.DataFrame, pandas.DataFrame, pandas.DataFrame): the most likely, upper, lower values
                Index
                    reset index
                Columns
                    predicted values (float)
        """
        predictor = _AutoTSPredictor(days=self._days, model_list="univariate")
        return predictor.predict(self._Y)

    def _multivariate_regression(self):
        """
        Perform multivariate forecasting (Regression with both of X(i-1) and Y(i-1)).

        Returns:
            tuple(pandas.DataFrame, pandas.DataFrame, pandas.DataFrame): the most likely, upper, lower values
                Index
                    reset index
                Columns
                    predicted values (float)
        """
        predictor = _AutoTSPredictor(days=self._days, model_list=["MultivariateRegression"])
        return predictor.predict(self._Y.join(self._X))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from matplotlib import pyplot as plt
from matplotlib.ticker import ScalarFormatter
import pandas as pd
from covsirphy.util.validator import Validator
from covsirphy.visualization.vbase import VisualizeBase, find_args


class LinePlot(VisualizeBase):
    """
    Create a line plot.

    Args:
        filename (str or None): filename to save the figure or None (display)
        bbox_inches (str): bounding box in inches when creating the figure
        kwargs: the other arguments of matplotlib.pyplot.savefig()
    """

    def __init__(self, filename=None, bbox_inches="tight", **kwargs):
        self._filename = filename
        self._savefig_dict = {"bbox_inches": bbox_inches, **kwargs}
        # Properties
        self._title = ""
        self._variables = []
        self._ax = None

    def __enter__(self):
        return super().__enter__()

    def __exit__(self, *exc_info):
        return super().__exit__(*exc_info)

    def plot(self, data, colormap=None, color_dict=None, **kwargs):
        """
        Plot chronological change of the data.

        Args:
            data (pandas.DataFrame or pandas.Series): data to show
                Index
                    x values
                Columns
                    y variables to show
            colormap (str, matplotlib colormap object or None): colormap, please refer to https://matplotlib.org/examples/color/colormaps_reference.html
            color_dict (dict[str, str] or None): dictionary of column names (keys) and colors (values)
            kwargs: keyword arguments of pandas.DataFrame.plot()
        """
        if isinstance(data, pd.Series):
            data = pd.DataFrame(data)
        Validator(data, "data").dataframe()
        self._variables = data.columns.tolist()
        # Color
        color_args = self._plot_colors(data.columns, colormap=colormap, color_dict=color_dict)
        # Set plotting
        try:
            self._ax = data.plot(**color_args, **kwargs)
        except ValueError as e:
            raise ValueError(e.args[0]) from None

    def x_axis(self, xlabel=None, x_logscale=False, xlim=(None, None)):
        """
        Set x axis.

        Args:
            xlabel (str or None): x-label
            x_logscale (bool): whether use log-scale in x-axis or not
            xlim (tuple(int or float, int or float)): limit of x domain

        Note:
            If None is included in xlim, the values will be automatically determined by Matplotlib
        """
        # Label
        self._ax.set_xlabel(xlabel)
        # Log scale
        if x_logscale:
            self._ax.set_xscale("log")
            xlim = (None, None) if xlim[0] == 0 else xlim
        # limit
        self._ax.set_xlim(*xlim)

    def y_axis(self, ylabel="Cases", y_logscale=False, ylim=(0, None), math_scale=True, y_integer=False):
        """
        Set x axis.

        Args:
            ylabel (str or None): y-label
            y_logscale (bool): whether use log-scale in y-axis or not
            ylim (tuple(int or float, int or float)): limit of y domain
            math_scale (bool): whether use LaTEX or not in y-label
            y_integer (bool): whether force to show the values as integer or not

        Note:
            If None is included in ylim, the values will be automatically determined by Matplotlib
        """
        # Label
        self._ax.set_ylabel(ylabel)
        # Math scale
        if math_scale:
            self._ax.yaxis.set_major_formatter(ScalarFormatter(useMathText=True))
            self._ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
        # Integer scale
        if y_integer:
            fmt = ScalarFormatter(useOffset=False)
            fmt.set_scientific(False)
            self._ax.yaxis.set_major_formatter(fmt)
        # Log scale
        if y_logscale:
            self._ax.set_yscale("log")
            ylim = (None, None) if ylim[0] == 0 else ylim
        # limit
        self._ax.set_ylim(*ylim)

    @staticmethod
    def _convert_to_list(x):
        """
        Convert None to empty list, str/float/int etc. to a list.

        Args:
            x (list/tuple[str/int/float] or None): value(s)

        Returns:
            list or tuple
        """
        return x if isinstance(x, (list, tuple)) else [] if x is None else [x]

    def line(self, v=None, h=None, color="black", linestyle=":"):
        """
        Show vertical/horizontal lines.

        Args:
            v (list/tuple[int/float] or None): list of x values of vertical lines or None
            h (list/tuple[int/float] or None): list of y values of horizontal lines or None
            color (str): color of the line
            linestyle (str): linestyle
        """
        # Horizontal
        h = self._convert_to_list(h)
        for value in h:
            self._ax.axhline(y=value, color=color, linestyle=linestyle)
        # Vertical
        v = self._convert_to_list(v)
        for value in v:
            self._ax.axvline(x=value, color=color, linestyle=linestyle)


def line_plot(df, title=None, filename=None, show_legend=True, **kwargs):
    """
    Wrapper function: show chronological change of the data.

    Args:
        data (pandas.DataFrame or pandas.Series): data to show
            Index
                Date (pandas.Timestamp)
            Columns
                variables to show
        title (str): title of the figure
        filename (str or None): filename to save the figure or None (display)
        show_legend (bool): whether show legend or not
        kwargs: keyword arguments of the following classes and methods.
            - covsirphy.LinePlot() and its methods,
            - matplotlib.pyplot.savefig(), matplotlib.pyplot.legend(),
            - pandas.DataFrame.plot()
    """
    with LinePlot(filename=filename, **find_args(plt.savefig, **kwargs)) as lp:
        lp.title = title
        lp.plot(data=df, **find_args([LinePlot.plot, pd.DataFrame.plot], **kwargs))
        # Axis
        lp.x_axis(**find_args([LinePlot.x_axis], **kwargs))
        lp.y_axis(**find_args([LinePlot.y_axis], **kwargs))
        # Vertical/horizontal lines
        lp.line(**find_args([LinePlot.line], **kwargs))
        # Legend
        if show_legend:
            lp.legend(**find_args([LinePlot.legend, plt.legend], **kwargs))
        else:
            lp.legend_hide()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
from datetime import datetime, timezone, timedelta
import os
from pathlib import Path
import urllib
import warnings
import datatable
import numpy as np
import pandas as pd
from unidecode import unidecode
from covsirphy.util.term import Term


class _DataProvider(Term):
    """Extract datasets and save it locally.

    Args:
        update_interval (int): update interval of downloading dataset
        stdout (str or None): stdout when downloading (shown at most one time) or None (no outputs)
    """

    def __init__(self, update_interval, stdout):
        self._update_interval = update_interval
        self._stdout = stdout

    def latest(self, filename, url, columns, date, date_format):
        """Provide the last dataset as a dataframe, downloading remote files or reading local files.

        Args:
            filename (str or pathlib.Path): filename to save/read the download
            url (str): URL of the dataset
            columns (list[str]): column names the dataset must have
            date (str or None): column name of date
            date_format (str): format of date column, like %Y-%m-%d

        Returns:
            pandas.DataFrame

        Note:
            If @verbose is 0, no descriptions will be shown.
            If @verbose is 1 or larger, URL and database name will be shown.
        """
        if not self.download_necessity(filename):
            with contextlib.suppress(ValueError):
                return self.read_csv(filename, columns, date=date, date_format=date_format)
        if self._stdout is not None:
            print(self._stdout)
            self._stdout = None
        df = self.read_csv(url, columns, date=date, date_format=date_format)
        df.to_csv(filename, index=False)
        return df

    @staticmethod
    def _last_updated_local(path):
        """
        Return the date last updated of local file/directory.

        Args:
            path (str or pathlib.Path): name of the file/directory

        Returns:
            datetime.datetime: time last updated (UTC)
        """
        m_time = Path(path).stat().st_mtime
        date = datetime.fromtimestamp(m_time)
        return date.astimezone(timezone.utc).replace(tzinfo=None)

    def download_necessity(self, filename):
        """
        Return whether we need to get the data from remote servers or not,
        comparing the last update of the files.

        Args:
            filename (str): filename of the local file

        Returns:
            (bool): whether we need to get the data from remote servers or not

        Note:
            If the last updated date is unknown, returns True.
            If @update_interval (of _DataProvider) hours have passed and the remote file was updated, return True.
        """
        if not Path(filename).exists():
            return True
        date_local = self._last_updated_local(filename)
        time_limit = date_local + timedelta(hours=self._update_interval)
        return datetime.now() >= time_limit

    @staticmethod
    def read_csv(path, columns, date, date_format):
        """Read the CSV file and return as a dataframe.

        Args:
            columns (list[str] or None): column names the dataset must have
            date (str or None): column name of date
            date_format (str): format of date column, like %Y-%m-%d

        Returns:
            pandas.DataFrame: downloaded data
        """
        try:
            with contextlib.redirect_stdout(open(os.devnull, "w")):
                df = datatable.fread(path, header=True).to_pandas()
            df = df.loc[:, columns or df.columns]
            for col in df:
                with contextlib.suppress(TypeError):
                    df[col] = df[col].apply(lambda x: unidecode(x) if len(x) else np.nan)
            if date is not None:
                df[date] = pd.to_datetime(df[date], format=date_format)
            return df.loc[:, columns or df.columns]
        except urllib.error.HTTPError:
            warnings.filterwarnings("ignore", category=pd.errors.DtypeWarning)
            kwargs = {
                "header": 0, "usecols": columns,
                "parse_dates": None if date is None else [date], "date_parser": lambda x: datetime.strptime(x, date_format)
            }
            return pd.read_csv(path, storage_options={"User-Agent": "Mozilla/5.0"}, **kwargs)
